import numpy as np
import os, re, sys
import matplotlib.pyplot as plt
from numba import njit, prange
from time import time

from .. import handle as hdl
from . import rotation as rot
from . import symmetries as sym
from ..odf import hsh
from ..odf import gridbased as grd
from .. import numba_plugins as nb
from ...config import data_type, hex_notation, odf_resolution

def parse_cif(file_path):
    """
    Parse a CIF file and calculate the positions of all atoms in the unit cell.
    
    Parameters:
        file_path (str): Path to the CIF file.

    Returns:
        dict: Contains lattice vectors, 
        atomic positions in fractional and Cartesian coordinates, and symmetry operations.
    """
    atom_labels = []
    atom_fractions = []
    space_group = None
    lattice_vectors = []
    symmetry_operations = []
    positions_cartesian = []
    occupancy = []
    B = []

    def parse_number(s):
        # this makes sure we can read numbers with given uncertainty, e.g. '4.9614(3)'
        return float(s.split('(')[0])

    try:
        with open(file_path, 'r') as f:
            lines = f.readlines()

        # Parse the CIF file line by line
        for i, line in enumerate(lines):
            # Extract space group
            if line.startswith('_space_group_IT_number'):
                space_group = int( line.split()[-1].strip("'\"") )
            
            # Extract lattice vectors
            if line.startswith('_cell_length_a'):
                a = parse_number(line.split()[-1])
            elif line.startswith('_cell_length_b'):
                b = parse_number(line.split()[-1])
            elif line.startswith('_cell_length_c'):
                c = parse_number(line.split()[-1])
            elif line.startswith('_cell_angle_alpha'):
                alpha = parse_number(line.split()[-1])
            elif line.startswith('_cell_angle_beta'):
                beta = parse_number(line.split()[-1])
            elif line.startswith('_cell_angle_gamma'):
                gamma = parse_number(line.split()[-1])

            # Find atomic data section
            if line.strip().startswith('loop_'):
                # Look for atomic position headers in the following lines
                headers = []
                j = i + 1
                while j < len(lines) and lines[j].strip().startswith('_'):
                    headers.append(lines[j].strip())
                    j += 1

                # Check if this block contains atomic positions
                if ('_atom_site_fract_x' in headers and
                        '_atom_site_fract_y' in headers and
                        '_atom_site_fract_z' in headers):
                    # Start reading data rows
                    while j < len(lines) and not lines[j].strip().startswith('loop_') and lines[j].strip():
                        split_line = lines[j].split()
                        atom_labels.append(split_line[headers.index('_atom_site_type_symbol')])  # Assume first column is atom type
                        atom_fractions.append([parse_number(split_line[headers.index('_atom_site_fract_x')]),
                                            parse_number(split_line[headers.index('_atom_site_fract_y')]),
                                            parse_number(split_line[headers.index('_atom_site_fract_z')])])
                        occupancy.append(parse_number(split_line[headers.index('_atom_site_occupancy')]))

                        # B.append(float(split_line[headers.index('_atom_site_U_iso_or_equiv')]))
                        j += 1

        # Convert lattice parameters to lattice vectors
        alpha, beta, gamma = np.radians([alpha, beta, gamma])
        lattice_vectors = [
            [a, 0, 0],
            [b * np.cos(gamma), b * np.sin(gamma), 0],
            [
                c * np.cos(beta),
                c * (np.cos(alpha) - np.cos(beta) * np.cos(gamma)) / np.sin(gamma),
                c * np.sqrt(1 - np.cos(beta) ** 2 - (np.cos(alpha) - np.cos(beta) * np.cos(gamma)) ** 2 / np.sin(gamma) ** 2),
            ],
        ]
        lattice_vectors = np.array(lattice_vectors)

        # Extract symmetry operations
        symmetry_mode = False
        for line in lines:
            if '_space_group_symop_operation_xyz' in line:
                symmetry_mode = True
            elif symmetry_mode and line.strip() == '':
                symmetry_mode = False
            elif symmetry_mode:
                op = line.strip().strip("'")
                symmetry_operations.append(op)

        # Apply symmetry operations to fractional coordinates
        atom_fractions = np.array(atom_fractions)
        full_fractions, atom_list, occupancy_list, B_list = [], [], [], []
        for op in symmetry_operations:
            # for (frac, at, oc, b) in zip(atom_fractions, atom_labels, occupancy, B):
            for (frac, at, oc) in zip(atom_fractions, atom_labels, occupancy):
                full_fractions.append(apply_symmetry_operation(op, frac))
                atom_list.append(at)
                occupancy_list.append(oc)
                # B_list.append(b)
        full_fractions = np.array(full_fractions)
        full_fractions = np.mod(full_fractions, 1)  # Ensure all coordinates are within [0, 1)
        full_fractions, indices = np.unique(np.squeeze(np.round(full_fractions,5)), axis=0, return_index=True)

        # Convert fractional to Cartesian coordinates
        positions_cartesian = np.array(
            [np.dot(frac, lattice_vectors) for frac in full_fractions]
        )
        # for frac in full_fractions:
        #     cart = np.dot(frac, lattice_vectors)
        #     positions_cartesian.append(cart)

        return {
            "atom_types": atom_labels,
            "coordinates": np.array(atom_fractions),
            "lattice_vectors": lattice_vectors,
            "space_group": space_group or "Unknown",
            'symmetry_operations': symmetry_operations,
            'atom_list': np.array(atom_list)[indices],
            'fractional_positions': full_fractions,
            'cartesian_positions': positions_cartesian,
            'occupancy': np.atleast_1d(occupancy_list)[indices],
            # 'B': np.atleast_1d(B_list)[indices],
        }

    except Exception as e:
        print(f"Error parsing CIF file: {e}")
        return None

def wavelength_from_E_keV(E_keV):
    h = 4.135667696e-18 # keV*s
    c = 299792458 # m/s
    wavelength = h*c*1e9 / E_keV # nm
    return wavelength

def get_reciprocal_space_coordinates(Qq_det, Chi_det, E_keV, geo):
    """Calculates the coordinates of the detector points based on
    polar coordinates q_det, chi_det and X-ray energy E_kev

    Returns
    -------
    ndarray
        3D reciprocal space coordinates and number of detector points
    """

    wavelength = wavelength_from_E_keV(E_keV)

    Two_theta = 2 * np.arcsin(Qq_det * wavelength / (4*np.pi))
    QX,QY,QZ = np.empty_like(Two_theta), np.empty_like(Two_theta), np.empty_like(Two_theta)
    for k in range(QY.shape[0]):
        QX[k], QY[k], QZ[k] = reciprocal_lattice_point(Two_theta[k], Chi_det[k], wavelength,
            u_beam=geo.beam_direction,
            u0=geo.detector_direction_origin,
            u90=geo.detector_direction_positive_90,
        )

    # import matplotlib.pyplot as plt
    # fig = plt.figure(figsize=(8, 6))
    # ax = fig.add_subplot(121, projection='3d')
    # ax.scatter(QX, QY, QZ)
    # ax.set_xlabel('x')
    # ax.set_ylabel('y')
    # ax.set_zlabel('z')
    return np.column_stack([QX, QY, QZ])

# @njit
def reciprocal_lattice_point(two_theta_rad, chi_rad, wavelength_nm, 
                             u_beam=(1,0,0), u0=(0,0,1), u90=(0,1,0)):
    """
    Calculate (qx, qy, qz) from angles and wavelength, defining the geometry by vectors.

    Parameters:
    - two_theta_rad: scattering angle (2?) in rad
    - chi_rad: azimuthal angle in rad
    - wavelength_nm: wavelength of incident beam in nm
    Optional:
    - u_beam: beam direction vector
    - u0: direction vector pointing towards the origin of chi
    - u90: direction vector pointing towards chi of +90 degree

    Returns:
    - (qx, qy, qz): reciprocal space coordinates (nm^-1)
    """
    # axes
    u_beam = np.array(u_beam, data_type)
    u0 = np.array(u0, data_type)
    u90 = np.array(u90, data_type)

    # Magnitude of wavevector
    k = 2 * np.pi / wavelength_nm

    # Incoming wavevector:
    k_in = k * u_beam

    # Outgoing wavevector (direction from spherical coordinates)
    k_out = k * (
        np.cos(two_theta_rad) * u_beam + 
        np.sin(two_theta_rad) * ( np.cos(chi_rad) * u0 + np.sin(chi_rad) * u90 )
        )

    q = k_out - k_in
    return q

def apply_symmetry_operation(operation, position):
    """
    Apply a symmetry operation to a fractional position.

    Parameters:
        operation (str): The symmetry operation in CIF format (e.g., "x,y,z").
        position (list or np.ndarray): Fractional coordinates [x, y, z].

    Returns:
        np.ndarray: New fractional coordinates after applying the operation.
    """
    # Ensure the symmetry operation format is consistent and parse it
    operation = operation.strip().replace(' ', '')  # Remove spaces for easier parsing
    operation = operation.replace('x', '{x}').replace('y', '{y}').replace('z', '{z}')

    # Check for simple transformations (e.g., -x or x+1/2)
    operation = re.sub(r'([+-]?\d*\.\d+|\d+)', r'float("\1")', operation)  # Convert numbers to float type

    x, y, z = position
    local_dict = {'x': x, 'y': y, 'z': z}

    # Safely evaluate the operation
    try:
        new_pos = [eval(operation.format(x=x, y=y, z=z))]
        return np.array(new_pos)
    except Exception as e:
        print(f"Error applying symmetry operation: {operation}")
        raise e

"""
array([[1.3268    , 3.70216157, 4.84805619],
       [3.9804    , 5.19063843, 1.57644381],
       [3.9804    , 8.14856157, 4.78869381],
       [1.3268    , 0.74423843, 1.63580619],
       [1.3268    , 6.72562464, 5.90861265],
       [3.9804    , 2.16717536, 0.51588735],
       [3.9804    , 2.27922464, 3.72813735],
       [1.3268    , 6.61357536, 2.69636265],
       [1.3268    , 8.0124128 , 5.85593175],
       [3.9804    , 0.8803872 , 0.56856825],
       [3.9804    , 3.5660128 , 3.78081825],
       [1.3268    , 5.3267872 , 2.64368175],
       [2.43918912, 6.08356448, 5.9066853 ],
       [2.86801088, 2.80923552, 0.5178147 ],
       [2.86801088, 1.63716448, 3.7300647 ],
       [2.43918912, 7.25563552, 2.6944353 ],
       [5.09278912, 2.80923552, 0.5178147 ],
       [0.21441088, 6.08356448, 5.9066853 ],
       [0.21441088, 7.25563552, 2.6944353 ],
       [5.09278912, 1.63716448, 3.7300647 ]])
       """

def get_reciprocal_points( lattice_vectors, hkl_list):
    """
    Calculate reciprocal lattice points for the given crystal up to the given miller indices.
    """
    # Direct lattice vectors 
    a1 = lattice_vectors[0]
    a2 = lattice_vectors[1]
    a3 = lattice_vectors[2]
    
    # Reciprocal lattice vectors
    V = np.dot(a1, np.cross(a2, a3))  # Unit cell volume
    b1 = 2 * np.pi * np.cross(a2, a3) / V
    b2 = 2 * np.pi * np.cross(a3, a1) / V
    b3 = 2 * np.pi * np.cross(a1, a2) / V
    
    # Generate reciprocal lattice points
    reciprocal_points = []
    for hkl in hkl_list:
        G = hkl[0] * b1 + hkl[1] * b2 + hkl[2] * b3
        reciprocal_points.append(G)
    
    return np.array(reciprocal_points)

def group_equivalent_reflections(hkl_max, lattice_vectors, point_group_quaternions):
    """Group reflections into families using symmetry operations."""
    families = {}
    
    # Generate reciprocal lattice points
    h_vals = np.arange(-hkl_max, hkl_max + 1)
    k_vals = np.arange(-hkl_max, hkl_max + 1)
    l_vals = np.arange(-hkl_max, hkl_max + 1)
    indices = []
    for h in h_vals:
        for k in k_vals:
            for l in l_vals:
                indices.append([h,k,l])

    # Reciprocal lattice vectors
    V = np.dot(lattice_vectors[0], np.cross(lattice_vectors[1], lattice_vectors[2]))  # Unit cell volume
    reciprocal_lattice_vectors = np.column_stack([
        2 * np.pi * np.cross(lattice_vectors[1], lattice_vectors[2]) / V,
        2 * np.pi * np.cross(lattice_vectors[2], lattice_vectors[0]) / V,
        2 * np.pi * np.cross(lattice_vectors[0], lattice_vectors[1]) / V
        ])

    for hkl in indices:
        # skip (0,0,0)
        if tuple(hkl) == (0,0,0):
            continue
        equivalents = equivalents_of_hkl(hkl, reciprocal_lattice_vectors, point_group_quaternions)

        # pick representative        
        pos_eq = np.all(np.greater_equal(equivalents,0),axis=1)
        if np.any(pos_eq):
            rep = tuple(max(np.array(equivalents)[pos_eq].tolist()))
        else:
            rep = max(equivalents)

        if rep not in families:
            families[rep] = set()
            families[rep].update(equivalents)
    
    representatives = []
    equivalents = []
    for rep, eqs in families.items():
        representatives.append(rep)
        equivalents.append(eqs)
    
    return representatives, equivalents


def equivalents_of_hkl(hkl, B, rotations, include_inversion=True,
                       tol=1e-4, friedel_equivalence=True):
    """
    Determine multiplicity of Miller index hkl under provided point-group rotations.
    Inputs:
      - hkl: tuple/list (h,k,l) integers
      - B: 3x3 reciprocal basis matrix [b1 b2 b3] (columns)
      - rotations: iterable of quaternions (proper rotations)
      - include_inversion: whether to also include inversion (-I)
      - tol: acceptance tolerance in reciprocal-space (same units as B)
      - friedel_equivalence: if True, treat hkl and -h-k-l as identical (Friedel)
    Returns:
      - unique_hkls: list of unique integer (h,k,l) generated
      - count: multiplicity (len of unique_hkls)
    """
    hkl = np.asarray(hkl, dtype=int)
    G = B @ hkl.reshape(3,1) 
    Binv = np.linalg.inv(B)
    found = []

    def accept_Gprime(Gp):
        # Convert back to h' (real)
        h_real = Binv @ Gp
        h_rounded = np.rint(h_real).astype(int).flatten()
        # compute residual in reciprocal space
        resid = np.linalg.norm(B @ h_rounded - Gp)
        if resid <= tol:
            return tuple(h_rounded)
        else:
            return None

    for q in rotations:
        Gp = rot.quaternion_rotate_vector( q, G.flatten() )
        h_ = accept_Gprime(Gp)
        if h_ is not None:
            found.append(h_)
        if include_inversion:
            Gp_inv = -Gp
            h2 = accept_Gprime(Gp_inv)
            if h2 is not None:
                found.append(h2)

    # canonicalize/fold Friedel pairs if requested
    unique_set = set()
    for h in found:
        if friedel_equivalence:
            h_neg = tuple((-np.array(h)).tolist())
            rep = max(h, h_neg)
        else:
            rep = h
        unique_set.add(rep)

    unique_hkls = sorted(unique_set)
    return unique_hkls

# def plot_powder_pattern(reciprocal_points, energy_keV, num_bins=100):
#     """Compute and plot the powder diffraction pattern (Intensity vs 2??)."""
#     wavelength = 12.398 / energy_keV  # Convert energy to wavelength (???)
#     G_magnitudes = np.linalg.norm(reciprocal_points, axis=1)  # Compute |G|
    
#     # Convert |G| to 2?? using Bragg's Law: 2?? = 2 * arcsin(??G / 4??)
#     theta_2 = 2 * np.arcsin(wavelength * G_magnitudes / (4 * np.pi)) * (180 / np.pi)  # Convert to degrees
    
#     # Create histogram (powder diffraction pattern)
#     hist, bin_edges = np.histogram(theta_2, bins=num_bins, density=True)
#     bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
    
#     # Plot powder pattern
#     plt.figure(figsize=(8, 5))
#     plt.plot(bin_centers, hist, '-k', lw=1.5)
#     plt.xlabel('2?? (degrees)')
#     plt.ylabel('Intensity (arb. units)')
#     plt.title('Simulated Powder Diffraction Pattern')
#     plt.grid(True)
#     plt.show()

#####
## there is the possibility to get the structure factor and others (Lorenz factor, solid angle correction, Debye-Waller factor, polarization correction)
## directly out of an average over the data
## Marc will not like this, but in many cases it might be more stable - it changes however the weights of the q-bins
# solidangle_factor = np.abs(np.sin(np.deg2rad(azim_angles)))[np.newaxis, np.newaxis, np.newaxis, :, np.newaxis] * np.ones(data_array.shape)
# intensity_scaling_factors = np.mean(data_array * solidangle_factor, axis = (0,1,2,3)) / np.mean(solidangle_factor, axis = (0,1,2,3))
# intensity_scaling_factors = intensity_scaling_factors/intensity_scaling_factors[0]
#####

# def structure_factor(hkl, lattice_vectors, frac_coords, species,
#                      form_factor_fn, occupancy=None, B=None):
#     """
#     hkl: tuple/list (h,k,l)
#     lattice_vectors: 3x3 array-like, columns = a, b, c (in angstroms)
#     frac_coords: Nx3 array of fractional coordinates (x,y,z)
#     species: list of length N with element symbols matching form_factor_fn
#     form_factor_fn: callable f(element, s) -> scalar real (atomic form factor)
#                     where s = sin(theta)/lambda
#     occupancy: optional list/array length N (default 1)
#     B: optional list/array length N of isotropic B factors (angstrom^2) (default 0)
#     Returns complex F and intensity |F|^2
#     """
#     h, k, l = hkl
#     A = np.asarray(lattice_vectors).reshape(3,3)
#     # compute G vector (reciprocal space)
#     A_inv = np.linalg.inv(A)
#     G = 2*np.pi * (A_inv.T @ np.array([h,k,l]))   # cartesian reciprocal vector
#     s = np.linalg.norm(G) / (4*np.pi)             # sin(theta)/lambda
#     coords = np.asarray(frac_coords)
#     N = coords.shape[0]
#     occ = np.ones(N) if occupancy is None else np.asarray(occupancy)
#     Barr = np.zeros(N) if B is None else np.asarray(B)
#     F = 0+0j
#     for j in range(N):
#         x,y,z = coords[j]
#         f = form_factor_fn(species[j], s)
#         phase = np.exp(2j*np.pi*(h*x + k*y + l*z))
#         thermal = np.exp(-Barr[j]*s*s)
#         F += occ[j] * f * thermal * phase
#     I = (F.real**2 + F.imag**2)
#     return F, I

def structure_factor_q(hkl, lattice_vectors, frac_coords,
                       form_factors, occupancy=None, B=None):
    """
    
    """
    h, k, l = hkl
    A = np.asarray(lattice_vectors).reshape(3,3)
    A_inv = np.linalg.inv(A)
    G = 2*np.pi * (A_inv.T @ np.array([h,k,l]))
    q = np.linalg.norm(G)
    coords = np.asarray(frac_coords)
    N = coords.shape[0]
    occ = np.ones(N) if occupancy is None else np.asarray(occupancy)
    Barr = np.zeros(N) if B is None else np.asarray(B)
    F = 0+0j
    for j in range(N):
        x,y,z = coords[j]
        f = form_factors[j]
        thermal = np.exp(-Barr[j]*(q/(4*np.pi))**2)
        phase = np.exp(2j*np.pi*(h*x + k*y + l*z))
        F += occ[j] * f * thermal * phase
    I = (F.real**2 + F.imag**2)

    return F, I

# @njit       
def structure_factor( reciprocal_points, atomic_form_factors, atom_positions, bool_pos_el ):
    '''
    reciprocal_points: 2D ndarray, float
        dim 0: reciprocal lattice points, 1: [qx,qy,qz]
    atomic_form_factors : 2D ndarray, float
        atomic form factors for different elements
        dim0: elements, dim1: q
    unit_cell_pos: 2D ndarray, float
        real space atomic positions
    bool_pos_el : 2D ndarray, string
        mask for uc_pos to find elements
        dim0: element dim1: mask for positions in uc

    Return values
    -------------
    SU_complex: 1D ndarray, complex128
        structure factor of the unit cell, function of Q
        dim0: complex structure factor
    '''
    S_complex=np.zeros_like(reciprocal_points[:,0], np.complex128)
    nEl=bool_pos_el.shape[0]
    
    for i in range(nEl): #for all elements 
        # f=atomic_form_factors[i] #get atomic form factor (for all Qs)
        S_element=np.zeros_like(S_complex)
        for r in atom_positions[bool_pos_el[i,:]]: #all atom positions of that element
            qr = np.dot(reciprocal_points,r)
            S_element += np.exp(qr*1j)
        S_complex += atomic_form_factors[i] * S_element
    return np.real(S_complex*S_complex.conjugate())

def structure_factor_from_cif(cif_path, 
            cutoff_structure_factor=1e-4, max_hkl=4, q_min=0, q_max=60, powder=False, upgrade_pointgroup=True):

    crystal_data = parse_cif(cif_path)
    lattice_vectors = crystal_data['lattice_vectors']/10 # assumes angstroem and converts to nm
    chem = crystal_data['atom_list'] # chemical symbols of all atoms in the unit cell
    atom_positions = crystal_data['cartesian_positions']/10 # coordinates of all atoms in the unit cell converted to nm
    space_group_no = crystal_data['space_group']
    # # alternatively import with ase:
    # import ase
    # unit = ase.io.read(cif_path)
    # atom_positions1 = unit.positions/10 # coordinates of all atoms converted to nm
    # chem = np.array( unit.get_chemical_symbols() ) # chemical symbols of all atoms
    # lattice_vectors = np.array( unit.get_cell() )/10 # converted to nm
    # space_group_no = unit.info['spacegroup'].no
    #
    if upgrade_pointgroup:
        symmetry = sym.get_diffraction_laue_group(space_group_no)
    else:
        symmetry = sym.get_proper_point_group(space_group_no)

    #### Calculate reciprocal lattice points
    gen = sym.generators(symmetry)
    Q_gen = rot.QfromOTP(gen)
    Q_group = rot.generate_group(Q_gen) # point group rotational symmetries
    hkl_list_full, hkl_equivalents_full = group_equivalent_reflections(max_hkl, lattice_vectors, Q_group)

    reciprocal_points_full = get_reciprocal_points( lattice_vectors, hkl_list_full )
    q_abs_full = np.linalg.norm(reciprocal_points_full, axis=1)
    # cut at max q
    mask = np.logical_and(q_abs_full < q_max, q_abs_full > q_min )
    hkl_list = np.array(hkl_list_full)[mask]
    hkl_equivalents = np.array(hkl_equivalents_full)[mask]
    reciprocal_points = reciprocal_points_full[mask]
    q_abs = q_abs_full[mask]

    #### Calculate atomic form factors
    elements = np.unique(chem)
    Nel = elements.size
    #bool mask for the unit cell
    atom_types_bool = np.empty( (Nel, chem.shape[0]), bool )
    for k in range(Nel):
        # determinate the positions of element k 
        atom_types_bool[k,:] = chem == elements[k]
    # load Cromer??Mann coefficients
    ff_path = hdl.get_file_path('textom',
        os.path.join('ressources','atomic_formfactors.txt'))
    ffcoeff = np.genfromtxt( ff_path, dtype="|U8", skip_header=2 )
    ffcoeff_used = np.atleast_2d([ ffcoeff[ np.where( el == ffcoeff )[0][0], 1: ] for el in elements]).astype(data_type)
    a = ffcoeff_used[:,0:-1:2]
    b = ffcoeff_used[:,1:-1:2]
    c = ffcoeff_used[:,-1]
    FF_element = np.empty( (Nel, q_abs.size), np.float64 ) # atomic form factors for all used elements
    for k in range(Nel):
        A, Qf = np.meshgrid( a[k], q_abs/10 ) # using angstroems^-1 here for simplicity
        B, _ = np.meshgrid( b[k], q_abs/10 )
        C = c[k]
        #Calculate form factor http://lampx.tugraz.at/~hadley/ss1/crystaldiffraction/atomicformfactors/formfactors.php
        F = ( A * np.exp( -B * (Qf/(4*np.pi))**2  ) ).sum(axis=1) + C
        # _, FF = np.meshgrid( self.Chi_det.reshape(self.detShape)[0], F )
        FF_element[k,:] = F
    #
    # plot to verify form factors:
    # idcs = np.argsort(q_abs)
    # plt.plot(q_abs[idcs],FF_element[0,idcs])
    # #

    ### Calculate structure factor from unit cell
    form_factors = np.array([FF_element[np.where(e==elements)[0][0]] for e in chem])
    structure_factors = np.empty_like(form_factors[0])
    for k in range(hkl_list.shape[0]):
        F,I=structure_factor_q(
            hkl_list[k], 
            crystal_data['lattice_vectors'].T, # note that it takes input in angstroem
            crystal_data['fractional_positions'], 
            form_factors[:,k], crystal_data['occupancy'], None )#crystal_data['B'])
        structure_factors[k] = I

    # structure_factors = structure_factor( reciprocal_points, FF_element, atom_positions, atom_types_bool)
    # # plot to verify structure factors:
    # idcs = np.argsort(q_abs)
    # plt.plot(q_abs[idcs],structure_factors[idcs], 'x')
    # #

    # exclude non-diffracting peaks from here (cut-off - corresponds to extinction conditions)
    mask_diffracting_peaks = structure_factors > cutoff_structure_factor*structure_factors.max()
    mask_diffracting_peaks[0] = False # exclude [0 0 0]
    q_used = q_abs[mask_diffracting_peaks]
    hkl_used = hkl_list[mask_diffracting_peaks]
    structure_factors_used = structure_factors[mask_diffracting_peaks]
    hkl_equivalents_used = hkl_equivalents[mask_diffracting_peaks]
    multiplicities_used = np.array([len(eq) for eq in hkl_equivalents_used])
    reciprocal_points_used = reciprocal_points[mask_diffracting_peaks]
    # and sort
    order = np.argsort(q_used)
    q_used = q_used[order]
    hkl_used = hkl_used[order]
    structure_factors_used = structure_factors_used[order]
    multiplicities_used = multiplicities_used[order] * 2 # for the friedel pairs
    hkl_equivalents_used = hkl_equivalents_used[order] # not returned atm
    reciprocal_points_used = reciprocal_points_used[order]

    reciprocal_points_used_full = np.concatenate([
        get_reciprocal_points(lattice_vectors, np.array(list(hkl_eq))) for hkl_eq in hkl_equivalents_used]
    )

    if hex_notation and symmetry[0] == '6':
        hkl_used = sym.miller_to_hex(hkl_used)

    if powder:
        return q_used, hkl_used, structure_factors_used, multiplicities_used, symmetry
    else:
        return q_used, reciprocal_points_used_full, structure_factors_used, multiplicities_used, hkl_used, symmetry
    # schreit nach einem structure_dict oder object

# @njit
# def quaternion_to_match_vectors(u, v):
#     """
#     Returns the quaternion (w, x, y, z) that rotates unit vector u into v.
#     u and v must be 3D unit vectors.
#     """
#     u = np.asarray(u, dtype=np.float64)
#     v = np.asarray(v, dtype=np.float64)
#     u /= np.linalg.norm(u)
#     v /= np.linalg.norm(v)

#     dot = np.dot(u, v)

#     if np.isclose(dot, 1.0):
#         # Vectors are the same
#         return np.array([1.0, 0.0, 0.0, 0.0])
#     elif np.isclose(dot, -1.0):
#         # Vectors are opposite
#         # Find an orthogonal vector
#         orthogonal = np.array([1.0, 0.0, 0.0])
#         if np.isclose(abs(u[0]), 1.0):
#             orthogonal = np.array([0.0, 1.0, 0.0])
#         axis = np.cross(u, orthogonal)
#         axis /= np.linalg.norm(axis)
#         return np.array([0.0, *axis])  # 180??? rotation: w=0
#     else:
#         axis = np.cross(u, v)
#         axis /= np.linalg.norm(axis)
#         angle = np.arccos(nb.nb_clip(dot, -1.0, 1.0))
#         w = np.cos(angle / 2.0)
#         xyz = axis * np.sin(angle / 2.0)
#         return np.concatenate(([w], xyz))

@njit
def quaternion_to_match_vectors(u, v):
    """
    Returns the quaternion (w, x, y, z) that rotates unit vector u into v.
    u and v must be 3D unit vectors.
    """
    u = u/nb.nb_vectornorm(u.astype(np.float64))
    v = v/nb.nb_vectornorm(v.astype(np.float64))

    d = nb.nb_dot(u, v)

    # Same direction
    if abs(d - 1.0) < 1e-12:
        return np.array([1.0, 0.0, 0.0, 0.0])

    # Opposite direction
    elif abs(d + 1.0) < 1e-12:
        # Find an orthogonal axis
        if abs(u[0]) < 0.9:
            ortho = np.array([1.0, 0.0, 0.0])
        else:
            ortho = np.array([0.0, 1.0, 0.0])
        axis = nb.nb_cross(u, ortho)
        axis = axis/nb.nb_vectornorm(axis)
        return np.array([0.0, axis[0], axis[1], axis[2]])

    # General case
    else:
        axis = nb.nb_cross(u, v)
        axis = axis/nb.nb_vectornorm(axis)
        # Clip dot to [-1,1] manually to avoid NaNs
        dd = d
        if dd > 1.0: dd = 1.0
        elif dd < -1.0: dd = -1.0
        angle = np.arccos(dd)
        w = np.cos(angle/2.0)
        s = np.sin(angle/2.0)
        return np.array([w, axis[0]*s, axis[1]*s, axis[2]*s])

@njit
def get_orientations_seen_on_pixel(q_detector, q_peak, sampling):
    '''
    q_peak is the reciprocal coordinate of the bragg-peak connected to no crystal rotation
        (center of the ODF)
    q_detector is the detector point reciprocal coordinate
    sampling is the number of rotation angles around the axis that are calculated
    '''

    if np.abs(nb.nb_vectornorm(q_detector)-nb.nb_vectornorm(q_peak)) > 1e-3: # 1e-10 is maybe a bit strict
        print("Can't rotate peak into desired q - check input")
        return # np.nan
    
    # Find a rotation that rotates the bragg peak on the detector pixel
    R_align = quaternion_to_match_vectors( q_peak, q_detector )

    # Define quaternion orientations that will produce the same bragg peaks
    axis = q_detector/nb.nb_vectornorm(q_detector) # i,j,k parts of Rot2
    ome = np.linspace(-np.pi, np.pi, sampling)
    w = np.cos(ome/2)
    ijk = np.zeros((sampling, 3)) 
    for i in range(3):
        ijk[:,i] = np.sin(ome/2) * axis[i]
    R_sampling = np.hstack((np.atleast_2d(w).T , ijk))
    # # invert all quaternions with negative first component (doesn't happen)
    # R_sampling[R_sampling[:,0]<0] = -R_sampling[R_sampling[:,0]<0]

    Rotations = np.zeros_like(R_sampling)
    for i in range(sampling):
        Rotations[i] = rot.quaternion_multiply( R_sampling[i], R_align)
        # xxxxx Rotations[i] = rot.quaternion_multiply( R_align, R_sampling[i] )
    
    return Rotations

@njit
def get_probability(q_peak, q_detector, odf_basis_function, odf_args, sampling=300):
    ''' 
    Paramters:
    --------
    q_peak: 1d ndarray, float 
        peak location in reciprocal space (in base orientation)
    q_detector: 1d ndarray, float
        desired peak location after rotation
    orientation_distribution: function
        probability function
    odf_args : tuple
        input for the function

    Returns:
    -------------
    proba: float
        probability that a peak appears at q_detector, given the ODF(mu,sig/kap)
    '''

    Rotations = get_orientations_seen_on_pixel(q_detector, q_peak, sampling)

    # orot = Rotation(Rotations)
    # oori = Orientation(orot, symmetry=point_group)
    # orot_fz = oori < OrientationRegion(point_group)

    # fig = plt.figure()
    # ax = fig.add_subplot(111, projection='3d')
    # ax.scatter(Rotations[:,1],Rotations[:,2],Rotations[:,3])
    # ax.set_xlim((-1,1))
    # ax.set_ylim((-1,1))
    # ax.set_zlim((-1,1))

    # evaluate the basis function at these orientations
    Probabilities = odf_basis_function( Rotations, *odf_args )
    # plt.plot(np.linspace(-180,180,num=Probabilities.size),Probabilities)

    probability_integrated = nb.trapz( Probabilities, 2*np.pi/sampling )
    return probability_integrated

@njit
def diffractlet( detector_coordinates, reciprocal_points, structure_factors, multiplicities, dtype_odf,
                odf_basis_function, odf_args ):

    n_peaks = structure_factors.size
    odf = np.zeros( detector_coordinates.shape[:2], dtype_odf ) # need complex for HSH
    for i_peak in range(n_peaks):
        for i_chi, q_detector in enumerate(detector_coordinates[:,i_peak]): 
            # for q in Q_group:
            #     point = rot.quaternion_rotate_vector(q, reciprocal_points[i_peak].copy())
            #     odf[i_peak, i_chi] += get_probability( 
            #         point, q_detector.copy(), 
            #         odf_basis_function, odf_args, sampling=100)
            #     # # friedel partner 
            #     # odf[i_peak, i_chi] += get_probability( 
            #     #     -point, q_detector.copy(), 
            #     #     odf_basis_function, odf_args, sampling=100)
            half_multiplicities = multiplicities//2 # don't need to include friedel pairs (test for symmetries without implied inversion symmetry)
            for k in range(half_multiplicities[i_peak]):
                idx = half_multiplicities[:i_peak].sum() + k
                odf[i_chi, i_peak] += get_probability( 
                    reciprocal_points[idx].copy(), q_detector.copy(), 
                    odf_basis_function, odf_args, sampling=100)
        # # ### test good sampling
        # samplings = np.logspace(np.log(50),np.log(10000),num=50,base=np.exp(1)).astype(int)
        # probabilities = []
        # for s in samplings:
        #     probabilities.append(get_probability( # angular coordinate missing
        #         reciprocal_points[i_peak].copy(), q_detector.copy(), 
        #         odf_basis_function, odf_args, sampling=s ))
        # plt.semilogx( samplings, np.array(probabilities)/probabilities[-1] )
        # # 100 - 1% off
        # # 200 - 0.5% off
        # # 900 - 0.1% off
        
        # print(f'\n\tFinished peak Nr.{i_peak+1}/{n_peaks}')

    difflet = odf * structure_factors[np.newaxis,:]
    return difflet

@njit
def hsh_wrapper(orientations_q, n, l, m):
    """
    Vectorized version: orientations in OTP notation
    Returns a 1D complex128 array of results.
    """
    OTP = rot.OTPfromQ(orientations_q)
    N = orientations_q.shape[0]
    out = np.empty(N, dtype=np.complex128)
    for i in range(N):
        out[i] = hsh.Z_numba(OTP[i,0], OTP[i,1], OTP[i,2], n, l, m)
    return out

@njit(parallel=True)
def diffractlets_parallel_hsh( order, symmetrization_matrix, 
                                detector_coordinates, Q_sample_rotations,
                                reciprocal_points, multiplicities, structure_factors):
    
    n_difflets, n_hsh = symmetrization_matrix.shape

    lm = np.empty((n_hsh, 2), dtype=np.int64)
    idx = 0
    for l in range(order+1):
        for m in range(-l, l+1):
            lm[idx, 0] = l
            lm[idx, 1] = m
            idx += 1

    n_rotations = Q_sample_rotations.shape[0]
    Gs = rot.OTPfromQ(Q_sample_rotations)

    difflets_hsh = np.empty( (n_hsh, n_difflets, *detector_coordinates.shape[:2]), np.complex128 )
    for k in prange( n_hsh ):
        # calculate complex HSH diffractlet
        hsh_difflet = diffractlet( 
                detector_coordinates.astype(np.float64), 
                reciprocal_points.astype(np.float64), structure_factors.astype(np.float64), multiplicities,
                difflets_hsh.dtype,
                hsh_wrapper, (order,lm[k,0],lm[k,1]) )
        # symmetrize
        for i_shsh in range(n_difflets):
                difflets_hsh[k, i_shsh] = symmetrization_matrix[i_shsh, k] * hsh_difflet
    difflets = np.real( difflets_hsh.sum(axis=0) )

    difflets_rot = np.empty( (n_rotations, n_difflets, detector_coordinates.shape[0], detector_coordinates.shape[1]), 
                data_type )
    rotations_sHSH = hsh.Rs_n_stack( np.column_stack( (-Gs[:,0], Gs[:,1], Gs[:,2]) ).astype(np.float64),
                                    order, symmetrization_matrix )
    for g, Rs_g in enumerate(rotations_sHSH):
        # difflets_rot[g] = (Rs_g @ difflets.reshape((difflets.shape[0], difflets.shape[1]*difflets.shape[2]))).reshape(difflets.shape)
        difflets_rot[g] = rotate_HSH_difflets( Rs_g, difflets)

    return difflets_rot, difflets_hsh

@njit
def rotate_HSH_difflets(Rs_g, difflets):
    n, _ = Rs_g.shape
    _, m, l = difflets.shape
    C = np.zeros((n, m, l), Rs_g.dtype)
    for i in range(n):
        for j in range(m):
            s = np.zeros(l, Rs_g.dtype)
            for k in range(n):
                s += Rs_g[i, k] * difflets[k, j]
            C[i, j] = s
    return C

@njit(parallel=True)
def diffractlets_parallel_grid( resolution, Q_grid, Q_group, 
                                detector_coordinates, Q_sample_rotations,
                                reciprocal_points, multiplicities, structure_factors ):
    
    n_rotations = Q_sample_rotations.shape[0]
    n_difflets = Q_grid.shape[0]
    difflets = np.empty( (n_rotations, n_difflets, *detector_coordinates.shape[:2]), np.float64 )
    for g in prange(n_rotations):
        detector_coordinates_rot = np.empty_like(detector_coordinates)
        for c in range(detector_coordinates.shape[0]):
            for q in range(detector_coordinates.shape[1]):
                detector_coordinates_rot[c,q] = rot.quaternion_inverse_rotate_vector(
                                            Q_sample_rotations[g], detector_coordinates[c,q])
                
        for k in range( n_difflets ):
            difflets[g,k] = diffractlet( 
                    detector_coordinates_rot.astype(np.float64), 
                    reciprocal_points.astype(np.float64), structure_factors.astype(np.float64), multiplicities,            
                    difflets.dtype,
                    grd.gaussian_SO3, 
                    (Q_grid[k].astype(np.float64), resolution*np.pi/180 / 2, Q_group.astype(np.float64))                    
                    # grd.fisher_SO3, 
                    # (Q_grid[k].astype(np.float64), 10/(resolution*np.pi/180)**2, Q_group.astype(np.float64))    
                    )

    return np.real(difflets).astype(data_type)

def get_diffractlets( cr, chi_det, geo, Q_sample_rotations, hsh_max_order=4, grid_resolution=25, 
                     cutoff_structure_factor=1e-4, max_hkl=4, odf_mode='hsh' ):
    """aa

    Parameters
    ----------
    cr : _type_
        _description_
    chi_det : _type_
        _description_
    geo : _type_
        _description_
    sample_rotations : _type_
        _description_
    orders : list or int, optional
        HSH orders, by default 4
    resolution : int, optional
        _description_, by default 25
    cutoff_structure_factor : float, optional
        _description_, by default 1e-4
    mode : str, optional
        _description_, by default 'hsh'

    Returns
    -------
    _type_
        _description_
    """

    q_hkl, reciprocal_points_full, structure_factors, multiplicities, hkl, symmetry = structure_factor_from_cif( 
        cr.cifPath, q_max=cr.q_range[1], cutoff_structure_factor=cutoff_structure_factor, max_hkl=max_hkl,
        upgrade_pointgroup=cr.use_upgraded_point_group )
    
    try:
        if cr.use_structure_factors_from_data:
            structure_factors = np.ones_like(structure_factors)
    except:
        print('\tCalculated structure_factors')

    Qq_det, Chi_det, detShape = rot.qchi(q_hkl, chi_det)
    detector_coordinates = get_reciprocal_space_coordinates( Qq_det, Chi_det, cr.E_keV, geo )
    detector_coordinates = detector_coordinates.reshape((*detShape,3))
    # pg = getattr( orx.osym, sym.get_SFnotation( symmetry ) )
    powder_pattern = structure_factors * multiplicities
    Gc, dV, V_fz = rot.sample_fundamental_zone( dchi=odf_resolution*np.pi/180, 
                            sampling='cubochoric', symmetry=symmetry )
    
    print('\tRetrieved reciprocal space coordinates')
    t0 = time()
    if odf_mode=='hsh':
        odf_obj = hsh.odf(hsh_max_order, sym.get_ppg_notation(symmetry), Gc)
        # ns = hsh.get_orders(symmetry, hsh_max_order, info=False)
        print('\tCalculating diffractlets')
        difflets_n = []
        for order in odf_obj.orders:
            if order==0:
                difflets = np.array(np.full( (Q_sample_rotations.shape[0],1,*detShape), 
                                np.atleast_2d( structure_factors*multiplicities) ), data_type)
            else:
                # symmetrization_matrix, _,_ = hsh.get_symmetrization_matrix( np.atleast_1d(ns), symmetry )
                difflets,_ = diffractlets_parallel_hsh( order, odf_obj.symmetrization_matrix[str(order)],
                            detector_coordinates, Q_sample_rotations,
                            reciprocal_points_full, multiplicities, structure_factors)
                # n_difflets = symmetrization_matrix[str(order)].shape[0]
                # difflets = np.zeros( (n_difflets, *detector_coordinates.shape[:2]), np.complex128 )
                # k = 0
                # for l in range(order):
                #     for m in range(-l,l+1):
                #         # calculate complex HSH diffractlet
                #         hsh_difflet = diffractlet( detector_coordinates, reciprocal_points, structure_factors, 
                #                     hsh.Z, (order,l,m) )
                #         # symmetrize
                #         for i_shsh in range(n_difflets):
                #                 difflets[i_shsh] += symmetrization_matrix[str(order)][i_shsh, k] * hsh_difflet
                #         k+=1
                #         # print(f'\tfinished n={order}, l={l}, m={m}')
            difflets_n.append(difflets)
            print(f'\t\tfinished order {order}')
        
        difflets_full = np.concatenate(difflets_n, axis=1)

    elif odf_mode=='grid':
        # Q_grid = rot.get_sample_fundamental(
        #                 resolution,
        #                 point_group= pg,
        #                 method='cubochoric'
        #         ).data.astype(data_type)
        
        # # Q_grid = rot.QfromOTP(np.array([[0,0,0], [np.pi/2,np.pi/2,np.pi/2]]))
        # # resolution = 10
        # [[738,47,697]] with resolution 10:
            # [1., 0., 0., 0.]
            # [ 0.7361178 , -0.6768535 ,  0.        ,  0.        ],
            # [ 0.7361178 ,  0.        , -0.6768535 ,  0.        ],

        # generators = sym.generators(symmetry)
        # q_gen = rot.QfromOTP(generators)
        # Q_group = rot.generate_group(q_gen)
        Q_grid, Q_group = grd.setup_gridbased(symmetry, grid_resolution)
        difflets_full = diffractlets_parallel_grid( grid_resolution, Q_grid[[738,47,697]] , 
                    Q_group, detector_coordinates, Q_sample_rotations,
                    reciprocal_points_full, multiplicities, structure_factors, )
        # print(Q_grid)

    else:
        print('\n\tMode not recognized, choose hsh or grid\n')
        sys.exit(1)
    
    # apply Lorentz correction
    theta = np.arcsin(wavelength_from_E_keV(cr.E_keV) * Qq_det / (4*np.pi))
    Lorentz = 1/np.sin(2*theta)
    difflets_full = difflets_full * Lorentz.reshape(detShape)[np.newaxis,np.newaxis,:,:]

    print(f'\t\ttook {(time()-t0)/60:.2f} min')
    return Qq_det.reshape(detShape), Chi_det.reshape(detShape), detector_coordinates, \
                    hkl, structure_factors, difflets_full, powder_pattern, symmetry, odf_obj

def plot_diffractlet( Qq_det, Chi_det, hkl, difflet, q_bins=50, cmap='bwr', logscale=False, sym_cmap=True ):

    f = plt.figure(figsize=(12,6))
    ax1 = f.add_subplot(121)
    ax2 = f.add_subplot(122, projection='polar')

    chi = Chi_det[:,0] 
    q_peaks = Qq_det[0]

    if isinstance(q_bins,np.ndarray):
        q_plot = q_bins
    else:
        q_plot = np.linspace(q_peaks[0],q_peaks[-1],num=q_bins)
    difflet_plot = np.zeros((difflet.shape[0], q_plot.size))
    for k in range(q_peaks.size):
        l = np.argmin(np.abs(q_plot-q_peaks[k]))
        difflet_plot[:,l] += difflet[:,k]
    Q_plot, Chi_plot = np.meshgrid(q_plot, chi)

    from matplotlib.colors import LogNorm, TwoSlopeNorm
    if logscale:
        difflet_log = difflet_plot - difflet_plot.min() + 1e-3
        ax1.pcolormesh(Q_plot, Chi_plot* 180/np.pi, difflet_log, cmap=cmap, norm=LogNorm())
        im=ax2.pcolormesh(Chi_plot, Q_plot, difflet_log, cmap=cmap, norm=LogNorm())
    elif sym_cmap:
        ax1.pcolormesh(Q_plot, Chi_plot* 180/np.pi, difflet_plot, cmap=cmap, norm=TwoSlopeNorm(vcenter=0))
        im=ax2.pcolormesh(Chi_plot, Q_plot, difflet_plot, cmap=cmap, norm=TwoSlopeNorm(vcenter=0))
    else:
        ax1.pcolormesh(Q_plot, Chi_plot* 180/np.pi, difflet_plot, cmap=cmap )
        im=ax2.pcolormesh(Chi_plot, Q_plot, difflet_plot, cmap=cmap )

    ax1.set_xlabel('q [nm^-1]')
    ax1.set_ylabel('chi [degree]')
    ax_top = ax1.twiny()
    ax_top.set_xlim(ax1.get_xlim())  # sync limits
    ax_top.set_xticks(q_peaks, labels=hkl)
    ax_top.tick_params(axis='x', labelrotation=45) 
    # ax1.set_xticks(q_peaks, labels=hkl )
    f.colorbar(im)
    f.tight_layout()

def plot_single_crystal_pattern(cr, cutoff_structure_factor=1e-4, axis=0):
    "very simple plot, axis (either 0,1,2) defines the beam direction with respect to the crystal structure"
    q_hkl, reciprocal_points_full, structure_factors, multiplicities, hkl, symmetry = structure_factor_from_cif( 
        cr.cifPath, q_max=cr.q_range[1], cutoff_structure_factor=cutoff_structure_factor )

    def other_axes(n:int):
        if n==0:
            return (1,2)
        elif n==1:
            return (2,0)
        elif n==2:
            return (0,1)
    det_axes = other_axes(axis)

    peaks = []
    for point in reciprocal_points_full:
        if np.abs(point[axis]) < 1e-2:
            peaks.append((point[det_axes[0]], point[det_axes[1]]))
    peaks = np.array(peaks)
    plt.scatter(peaks[:,0], peaks[:,1])

@njit
def get_odf_bin_indices(Q_sample, Q_bins, Q_group):
    """ bins quaternion in a predefined grid based on the maximum dot product with the bins
    """
    idcs = np.zeros(Q_sample.shape[0], dtype=np.int32)
    for s, q_s in enumerate(Q_sample):
        dots_bin = np.zeros(Q_bins.shape[0])
        for b, q_b in enumerate(Q_bins):
            dots_eq = np.zeros(Q_group.shape[0])
            for g, q_g in enumerate(Q_group):
                q_probed = rot.quaternion_multiply(q_g,q_s)
                q_probed *= q_probed[0] # invert quaternions with first 
                dot = rot.quaternion_dot(q_probed, q_b)
                # idcs_eq[g] = dot
                dots_eq[g] = dot
            dots_bin[b] = np.max(dots_eq)
        idcs[s] = np.argmax(dots_bin)
            
    return idcs
                 
def calculate_experimental_odf_coverage(cr, geo, 
                chi_det=np.linspace(0,2*np.pi,num=120), sample_rotations=[0,0,0], 
                cutoff_structure_factor=1e-4 ):

    q_hkl, reciprocal_points_full, structure_factors, multiplicities, hkl, symmetry = structure_factor_from_cif( 
        cr.cifPath, q_max=cr.q_range[1], cutoff_structure_factor=cutoff_structure_factor )
    
    # chosen_peaks = np.arange( multiplicities.size )
    chosen_peaks = [0,1] # check indices: hkl

    Qq_det, Chi_det, detShape = rot.qchi(q_hkl, chi_det)
    detector_coordinates = get_reciprocal_space_coordinates( Qq_det, Chi_det, cr.E_keV, geo )
    detector_coordinates = detector_coordinates.reshape((*detShape,3))

    q_gen = rot.QfromOTP(sym.generators(symmetry))
    Q_group = rot.generate_group(q_gen)

    Gc, dV, V_fz = rot.sample_fundamental_zone( dchi=5*np.pi/120 )
    Q_bins = rot.QfromOTP(Gc)
    odf_coverage = np.zeros(Gc.shape[0])
    for q in chosen_peaks:
        odf_coverage_q = np.zeros(Gc.shape[0])
        for q_d in detector_coordinates[:,q]:
            for q_p in reciprocal_points_full[multiplicities[:q].sum():multiplicities[:q+1].sum()]:

                ori_detected = get_orientations_seen_on_pixel(q_d, q_p, sampling=100)
                idcs = get_odf_bin_indices(ori_detected, Q_bins, Q_group)
                odf_coverage_q[np.unique(idcs)] += 1
        
        # not 100% sure about this but probably we should only count an orientation once per ring
        odf_coverage += odf_coverage_q.astype(bool)
        # maybe double-cover by equivalent points is ok but not by angle bins

            # plot a single q-point
            # orx.plot_points_in_fz( Q_bins[np.unique(idcs)], symmetry )
        # plot a full azimuthal coverage of a q-vector
        # orx.plot_points_in_fz( Q_bins[np.where(odf_coverage_q)[0]], symmetry ) 
        # plot double-covered bins
        # orx.plot_points_in_fz( Q_bins[np.where(odf_coverage_q > 1)[0]], symmetry ) 

    # plot a full azimuthal coverage of all qs
    # orx.plot_points_in_fz( Q_bins[np.where(odf_coverage)[0]], symmetry )
    # orx.plot_points_in_fz( Q_bins[np.where(odf_coverage > 1)[0]], symmetry )

    stop=1

@njit(parallel=True)
def inversion_matrix_parallel(chosen_peaks, chi_det, Q_samplerot, Q_bins,
                              detector_coordinates, reciprocal_points_full, multiplicities,
                              Q_group):
    n_data_proj = chosen_peaks.size * chi_det.size
    n_proj = Q_samplerot.shape[0]
    n_parameters = Q_bins.shape[0]
    A = np.zeros( (n_proj, n_data_proj, n_parameters), np.int8 )
    for r in prange(Q_samplerot.shape[0]):
        quat_rot = Q_samplerot[r]
        k = 0
        for q in chosen_peaks:
            for q_d in detector_coordinates[:,q]:
                q_d_rot = rot.quaternion_inverse_rotate_vector( quat_rot, q_d )
                for q_p in reciprocal_points_full[multiplicities[:q].sum():multiplicities[:q+1].sum()]:
                    ori_detected = get_orientations_seen_on_pixel(q_d_rot, q_p, sampling=100)
                    idcs = get_odf_bin_indices(ori_detected, Q_bins, Q_group)
                    A[r, k, np.unique(idcs)] += 1
                k+=1
    return A

def get_inversion_matrix( cr, geo, chosen_peaks, d_rot=10, tilts=[0,15,30,45], 
                               chi_resolution=6, cutoff_structure_factor=1e-4, upgrade_pointgroup=True):

    # let's see for a single ODF
    q_hkl, reciprocal_points_full, structure_factors, multiplicities, hkl, symmetry = structure_factor_from_cif( 
        cr.cifPath, q_max=cr.q_range[1], cutoff_structure_factor=cutoff_structure_factor, upgrade_pointgroup=upgrade_pointgroup )

    # chosen_peaks = [0,1] # check indices: hkl

    Gc, dV, V_fz = rot.sample_fundamental_zone( dchi=chi_resolution*np.pi/180, 
                                        sampling='cubochoric', symmetry=symmetry )  
    Q_bins = rot.QfromOTP( Gc )  
    
    Omega, Kappa = rot.samplerotations_eq3D(d_rot, tilts)
    Q_ome = np.column_stack(
            (np.cos(Omega/2),np.outer(np.sin(Omega/2), np.array(geo.inner_axis))))
    Q_kap = np.column_stack(
            (np.cos(Kappa/2),np.outer(np.sin(Kappa/2), np.array(geo.outer_axis))))
    Q_samplerot = np.array([rot.quaternion_multiply(qk,qo) for qo,qk in zip(Q_ome,Q_kap)])
    Gs = rot.OTPfromQ(Q_samplerot)

    chi_det=np.linspace(0,2*np.pi,num=360//chi_resolution)
    Qq_det, Chi_det, detShape = rot.qchi(q_hkl, chi_det)
    detector_coordinates = get_reciprocal_space_coordinates( Qq_det, Chi_det, cr.E_keV, geo )
    detector_coordinates = detector_coordinates.reshape((*detShape,3))

    q_gen = rot.QfromOTP(sym.generators(symmetry))
    Q_group = rot.generate_group(q_gen)

    A = inversion_matrix_parallel(np.array(chosen_peaks), chi_det, Q_samplerot, Q_bins,
                              detector_coordinates, reciprocal_points_full, multiplicities//2,
                              Q_group)

    return np.concatenate(A)

    # n_data_points = len(chosen_peaks) * chi_det.size * Q_samplerot.shape[0]
    # n_parameters = Gc.shape[0]
    # A = np.zeros( (n_data_points, n_parameters), np.int8 )
    # k = 0
    # for quat_rot in Q_samplerot:
    #     for q in chosen_peaks:
    #         for q_d in detector_coordinates[:,q]:
    #             q_d_rot = rot.quaternion_inverse_rotate_vector( quat_rot, q_d )
    #             for q_p in reciprocal_points_full[multiplicities[:q].sum():multiplicities[:q+1].sum()]:
    #                 ori_detected = get_orientations_seen_on_pixel(q_d_rot, q_p, sampling=100)
    #                 idcs = get_odf_bin_indices(ori_detected, Q_bins, Q_group)
    #                 A[k, np.unique(idcs)] += 1

    #             k+=1

    # U, s, Vt = np.linalg.svd(A)

    # # Small singular values -> unstable directions
    # tol = 1e-12  # or relative threshold like s[0]*1e-6
    # unstable = np.where(s < tol)[0]

    # # Look at rows of Vt corresponding to unstable singular values
    # print("Unstable directions in coefficient space:")
    # print(Vt[unstable])
    
    # # The ratio (largest ??? smallest) is the condition number. 
    # # If this number is huge, your inversion will be numerically unstable: 
    # # tiny changes or noise in d will blow up in the solution c
    # condition_number = np.max(s) / np.min(s)
    
    # The smallest singular value tells you how close A is to being rank-deficient (losing information). 
    # If a singular value is exactly zero, that means some directions in coefficient space can??t be 
    # recovered from d ?? the system is underdetermined.




import os, glob
import fabio
import h5py
import pyFAI
from time import time, sleep
import numpy as np
from subprocess import Popen
import shutil
import re
import sys

from .. import handle as hdl
from .. import misc as msc
from .. import exceptions as exc
from ...config import data_type
from ...input import integration_parameters

def setup_integration( sample_dir:str, confirm=True ):
    """Opens the integration_parameters.py file and saves it to sample_dir after modifying
    If such a file already exists, makes a backup and opens this one.

    Parameters
    ----------
    sample_dir : str
        textom base directory

    Returns
    -------
    dict
        parameters for pyfai integration for textom
    """
    # read and edit integration_parameters.py
    intpar_path_sample = os.path.join(sample_dir,'integration_parameters.py')
    if confirm:
        # check if there is already an integration file
        if not os.path.isfile( intpar_path_sample ):
            intpar_path_module = hdl.get_file_path('textom',os.path.join('input','integration_parameters.py'))
            # hdl.open_with_editor(intpar_path_module) # take the one from the textom module
            shutil.copyfile(intpar_path_module, intpar_path_sample ) # copy to the sample directory
        else:
            msc.cp_add_dt(intpar_path_sample, sample_dir, now=False) # save the old version with modification date
        hdl.open_with_editor(intpar_path_sample) # edit and use the same file
    
    par = msc.import_module_from_path('integration_parameters', intpar_path_sample)
    print(f'\tSaved integration parameters to {sample_dir}') 

    generate_caked_detector_mask(sample_dir, par)
    return par

def generate_caked_detector_mask(sample_dir:str, par:integration_parameters):
    """Integrates the provided raw detector mask so that it can be used in textom

    Parameters
    ----------
    sample_dir : str
        textom base directory
    par : integration_parameters
        contains what is given in the integration_parameters.py module
    """
    os.makedirs(os.path.join( sample_dir, 'analysis'), exist_ok=True)
    path_mask = os.path.join( sample_dir, 'analysis', 'mask_detector_cake.h5' )
    if not os.path.isfile(path_mask):
        # get mask and save its azimutal integration for further analysis
        mask = fabio.open(par.mask_path).data
        ai = pyFAI.load(par.poni_path)
        mask_cake = ai.integrate2d(
            np.ones_like(mask), 
            par.npt_rad, 
            par.npt_azi, 
            radial_range = par.rad_range, 
            azimuth_range = par.azi_range, 
            unit=par.rad_unit,
            method = par.int_method, 
            correctSolidAngle = par.solidangle_correction, 
            mask = mask, 
            safe = False,
        )
        with h5py.File(path_mask, 'w') as hf:
            hf.create_dataset('mask_cake', data = mask_cake.intensity)

def update_filelist( sample_dir:str, par:integration_parameters, ignore=[] ):
    """Reads all raw data names, filters them as indicated and checks if files are already there

    Parameters
    ----------
    sample_dir : str
        textom base directory
    par : integration_parameters
        contains what is given in the integration_parameters.py module
    ignore : list of str
        list of names of datasets to be skipped

    Returns
    -------
    fid_in : h5py file object
        this is the overview h5 file
    filtered dataset : list
        list of strings containing all datasets to be integrated
    """
    # compile the pattern for name matching
    fid_in = h5py.File( par.path_in, 'r' )
    repattern = '^' + par.h5_proj_pattern.replace('*', '(.*?)') + '$'
    repattern = re.compile(repattern)

    # get all datasets that correspond to the pattern
    filtered_datasets = []
    # scan_no = [] # feature to label the projections directly (?)
    n_integrated, n_missing = 0,0
    for entry in fid_in.keys():
        # Check if the name matches the pattern
        match = repattern.match(entry)
        if match:
            # print(entry)
            # Check if the dataset's data is actually present on the disk
            try:
                h5path_data = f'{entry}/{par.h5_data_path}'
                dataset = fid_in[h5path_data]
                # Attempt to access the dataset's shape (raises Error if the data is missing)
                _ = dataset.shape
            except:
                # print(f"Data for {entry} is missing on the disk.")
                continue  # Skip this entry if data is missing
        
            if entry in ignore:
                continue

            # Check if the dataset has already been integrated
            todo = False
            if par.mode%2:
                out_path = os.path.join(sample_dir,'data_integrated_1d', entry.split('.')[0]+'_integrate_1d.h5')
                todo = np.logical_or( todo, not os.path.isfile(out_path) )
                n_missing += todo
                n_integrated += not todo
            if par.mode>1:
                out_path = os.path.join(sample_dir,'data_integrated', entry.split('.')[0]+'_integrate_2d.h5')
                todo = np.logical_or( todo, not os.path.isfile(out_path) )
                n_missing += todo
                n_integrated += not todo
            if todo:
                # add the dataset to the integration list
                filtered_datasets.append(entry)
                # try:
                #     scan_no.append(int(match.group(1)))
                # except:
                #     pass
                    # print(f"Failed to extract scan number from entry: {entry}")
    print(f'\tRetrieved filelist, {n_integrated}/{n_missing+n_integrated} done')
    return fid_in, filtered_datasets

def start_integration_parallel( sample_dir:str, confirm=True, ignore=[] ):
    """Starts integration on several CPUs, calls integration_launcher.py

    Parameters
    ----------
    sample_dir : str
        textom base directory
    """
    par = setup_integration(sample_dir, confirm=confirm)
    print('Starting parallel pyFAI integration')
    t0 = time()
    # start parallel integration in separate processes
    int_path = hdl.get_file_path('textom',os.path.join('src','integration','integration_launcher.py'))
    pids = []
    for k in range(par.n_tasks):
        command = [
            'taskset', '-c', 
            '%d-%d' % (k*par.cores_per_task, (k+1)*par.cores_per_task-1),
            'python', int_path, 
            '-k', '%d' % (k),
            '-d', '%s' % (sample_dir),
        ]
        p = Popen(command)
        pids.append(p)
    for p in pids: # wait for all to be finished
        p.wait()
    for p in pids: # cleanup (not really necessary)
        p.kill()
    print('Integrations finished, total time: %d s' % (time()-t0))

def start_integration_online( sample_dir:str, wait=5., confirm=True, ignore=[], parallel=False ):
    """Starts integration on one GPU or CPU, updates the file list after each
    integration and continues until there is no raw data left 

    Parameters
    ----------
    sample_dir : str
        textom base directory
    wait : float
        waits this amount of seconds after integrating the last file
        before checking for new ones
    """
    par = setup_integration(sample_dir, confirm=confirm)
    print('Starting online pyFAI integration')
    t0 = time()
    if not parallel:
        flat=None
        if isinstance( par.flatfield_correction, str ):
            flat = fabio.open(par.flatfield_correction).data

        mask = fabio.open(par.mask_path).data
        try:
            ai = pyFAI.load(par.poni_path)
        except IndexError:
            raise exc.PyFAIrelated('Check if file paths in integration_parameters.py and your .poni file are correct!')

    fid_in, filtered_datasets = update_filelist( sample_dir, par, ignore=ignore )
    while len(filtered_datasets) > 0:
        try:
            if parallel:
                int_path = hdl.get_file_path('textom',os.path.join('src','integration','integration_launcher.py'))                
                #for gpu just launch parallel processes that will be sent to the gpu
                print(f'\tIntegrating {filtered_datasets[:par.n_tasks]}')
                pids = []
                for k in range(par.n_tasks):
                    command = [
                        'python', int_path, 
                        '%d-%d' % (k*par.cores_per_task, (k+1)*par.cores_per_task-1),
                        '-k', f'{k}',
                        '-d', sample_dir,
                    ]
                    p = Popen(command)
                    pids.append(p)
                for p in pids: # wait for all to be finished
                    p.wait()
                for p in pids: # cleanup (not really necessary)
                    p.kill()
            else:
                print(f'\tIntegrating {filtered_datasets[0]}')
                flexible_integrator(sample_dir, fid_in, filtered_datasets[0], par, ai, flat, mask,
                                    average_frames=par.average_frames)
        except Exception as e: 
            print('\t\t'+e)
            print(f'\t\tIntegration failed, ignoring dataset {filtered_datasets[0]}')
            ignore.append(filtered_datasets[0])
        fid_in.close()
        if len(filtered_datasets) == 1:
            print('\twaiting...')
            sleep(wait)
        fid_in, filtered_datasets = update_filelist( sample_dir, par, ignore=ignore )
    print('Integrations finished, total time: %d s' % (time()-t0))

def start_test_integration( sample_dir:str, dset_no=0 ):
    par = setup_integration(sample_dir, confirm=True)
    par.mode = 3 # for integrating both 1D and 2D
    flat=None
    if isinstance( par.flatfield_correction, str ):
        flat = fabio.open(par.flatfield_correction).data
    mask = fabio.open(par.mask_path).data
    ai = pyFAI.load(par.poni_path)
    fid_in, filtered_datasets = update_filelist( sample_dir, par, ignore=[] )
    print(f'\tIntegrating {filtered_datasets[dset_no]}')
    flexible_integrator(sample_dir, fid_in, filtered_datasets[dset_no], par, ai, flat, mask,
                        pick_center=True)
    # add "_test" to filenames
    list_of_int_files = glob.glob(os.path.join(sample_dir,'data_integrated','*.h5'))
    latest_file = max(list_of_int_files, key=os.path.getctime)
    new_name = os.path.splitext(latest_file)[0]+'_test.h5'
    os.rename(latest_file, new_name)
        # add "_test" to filenames
    test_filenames = []
    for subdir in ['data_integrated','data_integrated_1d']:
        list_of_int_files = glob.glob(os.path.join(sample_dir,subdir,'*.h5'))
        latest_file = max(list_of_int_files, key=os.path.getctime)
        new_name = os.path.splitext(latest_file)[0]+'_test.h5'
        os.rename(latest_file, new_name)
        test_filenames.append(new_name)
    return test_filenames

def flexible_integrator(sample_dir:str, fid_in, dataset:str, par:integration_parameters, ai, flat, mask, 
                        pick_center=False, average_frames=1):
    """Lauches pyfai integration, used in both parallel and online mode,
    both for 1d or 2d integration. Writes data and metadata in sample_dir/data_integrated/

    Parameters
    ----------
    sample_dir : str
        textom base directory
    fid_in : h5py file object
        this is the overview h5 file
    dataset : str
        path to the dataset in the overview h5 file
    par : integration_parameters
        contains what is given in the integration_parameters.py module
    ai : AzimuthalIntegrator object
        pyfai integrator
    flat : fabio image
        flat field correction
    mask : ndarray, 2d
        raw detector mask
    pick_center : bool
        only integrate some data for speed
    average_frames : int
        made for continuous rotation mode to reduce angular sampling
    """
    # get paths for the correct h5 file
    h5path_data = f'{dataset}/{par.h5_data_path}'
    metadata = read_metadata(fid_in, dataset, par, average_frames=average_frames)

    #     out_name = '{}_{:03d}_{:03d}_{:.0f}_{:08.2f}_diff_scan_0001_comb'.format(
    #                 par.title,
    #                 scan_no[l],scan_no[l],
    #                 fid_in[h5path_tilt][()],
    #                 fid_in[h5path_rot][()],
    #             ).replace('.','p')
    # else:
    out_name = dataset.split('.')[0]

    data_in = fid_in[h5path_data]#[()]
    if pick_center:
        # print('yay')
        # data_in = np.atleast_3d( np.mean(
            # data_in[ metadata['npts_fast_scanaxis'] * metadata['npts_fast_scanaxis'] // 2 : metadata['npts_fast_scanaxis'] * ( metadata['npts_fast_scanaxis'] // 2 + 1 )],
            # axis=0 ) )
        try: #what am i trying to do here? does it work?
            data_in = data_in[ metadata['npts_slow_scanaxis'] * metadata['npts_fast_scanaxis'] // 2 : metadata['npts_slow_scanaxis'] * ( metadata['npts_fast_scanaxis'] // 2 + 1 )]
        except:
            pass
        # data_in = np.atleast_2d(np.max(data_in,axis=0))
        # print('yo')

    # average data if required
    if average_frames > 1:
        m = data_in.shape[0] // average_frames
        summed_frames = data_in[:m * average_frames].reshape((m, average_frames, *data_in.shape[1:])).sum(axis=1)
        data_in = summed_frames

    n_frames = data_in.shape[0]

    if par.mode%2:
        os.makedirs(os.path.join(sample_dir, 'data_integrated_1d/'), exist_ok=True)
        path_out = os.path.join(
            sample_dir, 'data_integrated_1d/',
            out_name + '_integrate_1d.h5'
        )
        fid_out = h5py.File( path_out, 'w' )
        write_metadata( metadata, fid_out )

        radial_dset = fid_out.create_dataset( 'radial_units', (1,par.npt_rad_1D) )
        intensity_dset = fid_out.create_dataset(
            'cake_integ',
            ( n_frames, par.npt_rad_1D ),
            chunks = ( 1, par.npt_rad_1D ),
            shuffle="True", compression="lzf",
            dtype=data_type,
            )

        t0 = time()
        for frame in range (0,n_frames):
            # print(frame)
                
            result1D = ai.integrate1d(
                data_in[frame,:,:], 
                par.npt_rad_1D, 
                radial_range = par.rad_range, 
                unit=par.rad_unit,
                method = par.int_method, 
                correctSolidAngle = par.solidangle_correction, 
                dark = par.darkcurrent_correction,
                flat = flat,
                mask = mask, 
                polarization_factor = par.polarisation_factor, 
                safe = False,
            )

            radial_dset[0,:] = result1D.radial
            intensity_dset[frame,:]= result1D.intensity

            t_it = (time()-t0)/(frame+1)
            sys.stdout.write(f'\r\t\tIntegrated frames: {(frame+1):d} / {n_frames:d}, t/frame: {t_it:.3f}' )
            sys.stdout.flush()
        # ai.reset()
        fid_out.close()
        print(f', total time: {time()-t0:.1f} s')

    if par.mode>1:
        os.makedirs(os.path.join(sample_dir, 'data_integrated/'), exist_ok=True)
        path_out = os.path.join(
            sample_dir, 'data_integrated/',
            out_name + '_integrate_2d.h5'
        )
        fid_out = h5py.File( path_out, 'w' )
        write_metadata(metadata, fid_out)

        radial_dset = fid_out.create_dataset( 'radial_units', (1,par.npt_rad) )
        azimuthal_dset = fid_out.create_dataset( 'azimuthal_units', (1,par.npt_azi) )
        intensity_dset = fid_out.create_dataset(
                'cake_integ',
                ( n_frames, par.npt_azi, par.npt_rad ),
                chunks = ( 1, par.npt_azi, par.npt_rad ),
                shuffle="True", compression="lzf",
                dtype=data_type,
                )
        
        t0 = time()
        for frame in range (0,n_frames):
            result2D = ai.integrate2d(
                data_in[frame,:,:], 
                par.npt_rad, 
                par.npt_azi, 
                radial_range = par.rad_range, 
                azimuth_range= par.azi_range, 
                unit=par.rad_unit,
                method = par.int_method, 
                correctSolidAngle = par.solidangle_correction, 
                dark = par.darkcurrent_correction,
                flat = flat,
                mask = mask, 
                polarization_factor = par.polarisation_factor, 
                safe = False,
            )

            radial_dset[0,:] = result2D.radial
            azimuthal_dset[0,:] = result2D.azimuthal
            intensity_dset[frame,:,:]= result2D.intensity

            t_it = (time()-t0)/(frame+1)
            sys.stdout.write(f'\r\t\tIntegrated frames: {(frame+1):d} / {n_frames:d}, t/frame: {t_it:.2f}' )
            sys.stdout.flush()
        # ai.reset()
        fid_out.close()
        print(f'Total time: {time()-t0:.1f} s')

def read_metadata( fid_in, dataset, par:integration_parameters, average_frames=1 ):
    """Reads metadata from raw file as given in the integration_parameters.py file

    Parameters
    ----------
    fid_in : h5py file object
        this is the overview h5 file
    dataset : str
        path to the dataset in the overview h5 file
    par : integration_parameters
        contains what is given in the integration_parameters.py module

    Returns
    -------
    dict
        metadata and names to be written in the integrated file

    Raises
    ------
    exc.metadata_missing
        aborts if required metadata are missing
    """
    metadata = {}

    # Read required metadata
    required = [
        ('rot_angle', par.h5_rot_angle_path),
        ('tilt_angle', par.h5_tilt_angle_path),
        ('ty', par.h5_ty_path),
        ('tz', par.h5_tz_path),
    ]
    for dset_name, dset_path in required:
        try:
            h5path = f'{dataset}/{dset_path}'
            metadata[dset_name] = fid_in[h5path][()]
        except:
            if dset_name == 'rot_angle':
                splnm = dataset.split('_')
                i_diff = next(i for i, s in enumerate(splnm) if 'diff' in s) # find where is 'diff'
                print('\t\tReading rotation angle from name')
                metadata[dset_name] = float( splnm[i_diff-1].replace('m','-').replace('p','.') ) 
            elif dset_name == 'tilt_angle':
                splnm = dataset.split('_')
                i_diff = next(i for i, s in enumerate(splnm) if 'diff' in s) # find where is 'diff'
                metadata[dset_name]= float( splnm[i_diff-2].replace('m','-').replace('p','.') )
                print('\t\tReading rotation angle from name')
            else:
                print(f'\t\tRequired metadata in {h5path} not found, stopping' )
                raise exc.metadata_missing(f'Required metadata in {h5path} not found')

    # this differs depending on scan or controt
        # ('npts_fast_scanaxis',par.h5_nfast_path),
        # ('npts_slow_scanaxis',par.h5_nslow_path),
    if par.h5_nfast_path:
        try:
            h5path_nfast = f'{dataset}/{par.h5_nfast_path}'
            nfast = fid_in[h5path_nfast][()]
        # except:
        #     print(f'\t\tRequired metadata in {h5path_nfast} not found, stopping' )
        #     raise exc.metadata_missing(f'Required metadata in {h5path_nfast} not found')
        # try:
            h5path_nslow = f'{dataset}/{par.h5_nslow_path}'
            nslow = fid_in[h5path_nslow][()]
        except:
            # print(f'\t\tRequired metadata in {h5path_nslow} not found, stopping' )
            # raise exc.metadata_missing(f'Required metadata in {h5path_nslow} not found')
            nfast, nslow = -1, -1
        metadata['fov'] = ( nfast, nslow )
        metadata['npts_fast_scanaxis'] = nfast
        metadata['npts_slow_scanaxis'] = nslow

    # read optional metadata
    optional = [
        ('ion', par.h5_ion_path),
        ('data_transmission', par.h5_data_transmission_path),
    ]
    for dset_name, dset_path in optional:
        try:
            if dset_path:
                h5path = f'{dataset}/{dset_path}'
                metadata[dset_name] = fid_in[h5path][()]
            else:
                pass
                # metadata[dset_name] = None
        except:
            print(f'\t\tOptional metadata in {h5path} not found, continuing' )

    # average frames if required
    for dset_name in metadata.keys():
        if len(np.atleast_1d(metadata[dset_name])) > 1 and average_frames > 1:
            m = len(metadata[dset_name]) // average_frames
            averaged = metadata[dset_name][:m * average_frames].reshape(m, average_frames).mean(axis=1)
            metadata[dset_name] = averaged
            
    return metadata

def write_metadata( metadata, fid_out ):
    """Writes metadata to the integrated data file

    Parameters
    ----------
    metadata : dict
        metadata and names to be written in the integrated file
    fid_out : h5py file object
        this is the integrated data h5 file
    """
    for key in metadata:
        fid_out.create_dataset( key, data = metadata[key])

###############################################################################################
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib.widgets import Button
from numba import njit, prange
from ..misc import import_module_from_path

def check_detector_geometry( intpar_path:str, testfile:str, testfile_dataset:str, frame_no:int, geo_path:str, 
                            vmin=1,vmax = 10, logscale = False ):



    par = msc.import_module_from_path('integration_parameters', intpar_path)
    mask = fabio.open(par.mask_path).data
    ai = pyFAI.load(par.poni_path)

    # integrate the test-dataset
    fid_in = h5py.File(testfile, 'r')
    # test_data = np.sum( fid_in['2.1'][par.h5_data_path],axis=0 )#[0] # check if 0 is ok (can pass it also as argument or plot several ?)
    test_data = fid_in[testfile_dataset][par.h5_data_path][frame_no] # check if 0 is ok (can pass it also as argument or plot several ?)
    result2D = ai.integrate2d(
                test_data, 
                par.npt_rad, 
                par.npt_azi, 
                radial_range = par.rad_range, 
                azimuth_range= par.azi_range, 
                unit=par.rad_unit,
                method = par.int_method, 
                correctSolidAngle = par.solidangle_correction, 
                dark = par.darkcurrent_correction,
                mask = mask, 
                polarization_factor = par.polarisation_factor, 
                safe = False,
            )
    q = result2D.radial
    chi = result2D.azimuthal * np.pi/180
    test_data_polar = result2D.intensity

    # test_data_polar[60,-1] =test_data_polar.max()
    # test_data_polar[60,-2] =test_data_polar.max()
    # test_data_polar[60,-3] =test_data_polar.max()
    # test_data_polar[90,-1] =np.nan
    # test_data_polar[90,-2] =np.nan
    # test_data_polar[90,-3] =np.nan
    geo = import_module_from_path('geometry', geo_path )
    det_Y,det_Z = get_detector_coordinates(geo, q, chi)

    fig = plt.figure(figsize=(18, 10))
    ax1 = fig.add_subplot(121)
    ax1.matshow( test_data, vmin=vmin,vmax=vmax, 
                norm=LogNorm() if logscale else None,
                )
    if geo.flip_detector_lr:
        ax1.invert_xaxis()
    if geo.flip_detector_ud:
        ax1.invert_yaxis()

    ax2 = fig.add_subplot(122)#, polar=True)
    ax2.pcolormesh( det_Y, det_Z, test_data_polar, vmax=vmax,
                   norm=LogNorm() if logscale else None,
                   )
    # ax2.set_axis_off()
    ax1.set_title('Raw data')
    ax1.set_xlabel(r'$\leftarrow y$')
    ax1.set_ylabel(r'$z \rightarrow$')
    ax2.set_xlabel(f'$q_y$ [nm$^-1$]')
    ax2.set_ylabel(f'$q_z$ [nm$^-1$]')
    ax2.invert_xaxis()
    ax2.set_title('Integrated data')
    plt.axis('equal')
    plt.show(block=True)

def get_detector_coordinates(geo, q, chi):
    # load geometry and calculate qx, qy
    # CHi, QQ = np.meshgrid( chi, q )
    QQ, CHi = np.meshgrid( q, chi )
    # Y = -QQ*np.sin(CHi)
    # Z = QQ*np.cos(CHi)
    Y,Z = np.empty_like(CHi), np.empty_like(CHi)
    for k in range(Y.shape[0]):
        for l in range(Y.shape[1]):
            Y[k,l], Z[k,l] = polar_to_cartesian(QQ[k,l], CHi[k,l], 
                u0=(geo.detector_direction_origin[1], geo.detector_direction_origin[2]),
                u90=(geo.detector_direction_positive_90[1], geo.detector_direction_positive_90[2]) # 
            )
    return Y, Z

@njit
def polar_to_cartesian(q:float, chi:float, u0=(0,1), u90=(1,0)):
    # Ensure numpy arrays
    u0 = np.array(u0)
    u90 = np.array(u90)
    
    # Compute Cartesian coordinates
    point = q * (np.cos(chi) * u0 + np.sin(chi) * u90)
    return point[0], point[1]

def flatmesh( x, y):
    XX,YY = np.meshgrid(x,y)
    return XX.flatten(), YY.flatten()


import numpy as np
import textom.src.model.rotation as rot

# Sample rotations
angular_stepsize_vertical_rotation = 30 # degree
tilt_angles = (0,30,60)
omega, kappa = rot.samplerotations_eq3D(d_rot=angular_stepsize_vertical_rotation, tilts=tilt_angles)

# Sample size and shape
sample_outline_size = (4,5,6) # no of voxels in (x,y,z)
buffer = 1 # number of zero-voxels around the sample
sample_mask = np.zeros( sample_outline_size, bool )
sample_mask[buffer:-buffer,buffer:-buffer,buffer:-buffer] = 1

# detector points
q_det = np.linspace( 10, 35, num=100 )
chi_det = np.linspace(0, 360, num=120, endpoint=False)

# odf coefficients
intensity_style = 'uniform' # ...
coefficient_style = 'single_first' # 'single_random' #...
load_file = False # enable to load the parameters from a path


import numpy as np
import matplotlib.pyplot as plt

def generate_hemisphere_points(num_points, y_max=1.0):
    """
    Generate evenly distributed points on the surface of a hemisphere.

    Parameters:
        num_points (int): Number of points to generate.
        y_max (float): Maximum y-coordinate value (between 0 and 1).

    Returns:
        np.ndarray: Array of shape (num_points, 3) containing the 3D coordinates of the points.
    """
    if not (0 < y_max <= 1.0):
        raise ValueError("y_max must be between 0 and 1.")

    points = []
    golden_angle = np.pi * (3 - np.sqrt(5))  # Golden angle in radians

    for i in range(num_points):
        y = i / (num_points - 1) * y_max  # y-coordinate ranges from 0 to y_max for a hemisphere
        radius = np.sqrt(1 - y**2)  # Radius of the circle at height y

        theta = i * golden_angle  # Angle around the circle
        x = radius * np.cos(theta)
        z = radius * np.sin(theta)

        points.append([x, y, z])

    return np.array(points)

# Example usage
num_points = 100
y_max = 0.8
hemisphere_points = generate_hemisphere_points(num_points, y_max=y_max)

# Plot the points in 3D
fig = plt.figure(figsize=(8, 6))
ax = fig.add_subplot(111, projection='3d')
ax.scatter(hemisphere_points[:, 0], hemisphere_points[:, 1], hemisphere_points[:, 2], c='b', marker='o')
ax.set_xlabel('X')
ax.set_ylabel('Y')
ax.set_zlabel('Z')
ax.set_title('Points on a Hemisphere with y_max = {:.2f}'.format(y_max))
plt.show()


import os
# Housekeeping, set num threads to 1 before importing numpy
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1"
import numpy as np
import matplotlib.pyplot as plt
import pickle

import src.handle as hdl
from textom.src.model.model_textom import model_textom
from textom.src.data_treatment.data import data
from textom.src.optimization.fitting import fitting
import textom.src.plotting as myplt
from textom.src.optimization.optimizer import optimize
import textom.src.model.rotation as rot
import textom.src.model.symmetries as sym
from config import pickle_path

# this is to avoid an error on greenland
import matplotlib 
matplotlib.use('TkAgg')

#################### INPUT ##################

mod = model_textom( 'input.input_simsample' )
projections = 'full' #'notilt'
n_max = 12 # maximum order to optimize

#############################################

with open(os.path.join(pickle_path, 'data_%s.txt' % mod.title),'rb') as fid:
    dat = pickle.load(fid)
fit = fitting(dat,mod)
qmask = np.array([False, False, False, False, False, False, False,  True,  True,
        True, False, False,  True,  True,  True,  True, False, False,
        True,  True, False,  True, False, False, False, False, False,
        True,  True, False, False, False, False, False, False, False,
       False,  True,  True,  True, False,  True, False, False,  True,
       False,  True,  True, False, False])
fit.mask_q( mask = qmask )
hdl.save_pickle('fit_%s' % mod.title, fit, 'fit')

fit.choose_projections( projections ) 

loss_all, MDL = [], []
print( 'Optimize 0 and lowest order to get orienations')
fit.C, opt0 = optimize( fit, 0 ) # opimize c0 from average intensity
fit.set_orders( mod, 4 )
fit.C, opt = optimize( fit, 1 ) # optimize only 4th order
loss_all.append(opt['loss'])
fit.C, opt = optimize( fit, 2, tol=1e-3 ) # optimize all together
loss_all.append(opt['loss'])
MDL.append( fit.MDL(opt['loss'][-1]) )
c_res = fit.C.copy()

# make figure that shows the deviation from the real mean direction
g_pref = mod.g_ML_sample( fit.C[mod.mask_voxels] )
file = hdl.read_vtk( 'output/sim_%s.vtk' % mod.title )
mu = file[1]['mu'][mod.mask_voxels]
gen = sym.generators(mod.symmetry)
dis = rot.ang_distance(rot.QfromOTP(g_pref),rot.QfromOTP(mu),gen)
plt.figure()
plt.hist(dis*180/np.pi)
plt.ylabel('No of voxels')
plt.xlabel('dg [%s]' % chr(176))

print( 'Optimize higher orders for full ODF' )
for n in range(6,n_max+1,2):
    fit.set_orders(mod,n)
    fit.C, opt = optimize( fit, 1 ) # optimize only n-th order
    loss_all.append(opt['loss'])
    fit.C, opt = optimize( fit, 2, tol=1e-3 ) # optimize all together
    loss_all.append(opt['loss'])
    MDL.append( fit.MDL(opt['loss'][-1]) )
    if np.min(MDL) == MDL[-1]:
        c_res = fit.C.copy()

# plot the development of the loss function
plt.figure()
t0 = 0
for k, loss in enumerate(loss_all):
    plt.plot(range(t0,t0+len(loss)), loss, 
             label='n = %d' % fit.ns[int(k/2)+1])
    t0 += len(loss)
plt.legend()
plt.xlabel('Iterations')
plt.ylabel('Loss function')

# plot the minimum description length
plt.figure()
plt.plot(fit.ns[1:],MDL)
plt.xlabel('n')
plt.ylabel('MDL')

# get variances using the result with min(MDL)
n_chosen = fit.ns[np.argmin(MDL) + 1]
mod.set_orders(n_chosen)
stds = mod.std_sample( c_res[mod.mask_voxels], g_pref )
plt.figure()
plt.hist(stds *180/np.pi, bins=20)
plt.xlabel('%s [%s]' % (chr(963),chr(176)))
plt.ylabel('No of Voxels')

# export vtk-file for paraview
a_pref, b_pref, c_pref = mod.abc_pref_sample( g_pref )
path = 'output/fit_%s.vtk' % mod.title
hdl.write_vtk_direct( mod, path, mod.mask_voxels,
                     scalars = [
                        ['scaling', c_res[mod.mask_voxels,0]],
                        ['std', stds ]
                      ],
                     vectors = [
                        ['a_pref', a_pref],
                        ['b_pref', b_pref],
                        ['c_pref', c_pref],
                        ]
                    )
print('Saved mean orientations to file: %s' % path)

plt.show()


import os, re
import numpy as np
from time import time
import h5py

# domestic
from ..numba_plugins import sparsemult, integrate_c, masked_mean_axis0
from ..misc import check_version
from ...config import data_type, hex_notation
from ..model.model_textom import model_textom
from ..model.symmetries import miller_to_hex
from ..data_treatment.data import load_peak_regions
from ..odf import hsh
from ..exceptions import dimension_error

from numba import njit, prange
import warnings
from numba.core.errors import NumbaDeprecationWarning, NumbaPerformanceWarning
warnings.simplefilter('ignore', category=NumbaPerformanceWarning)

class fitting:
    """ A class to that contains loss function and gradients for fitting
    texture tomography data

    Takes input from a model and a data class object or a previously saved 'fit' pickle
    """
    def __init__(self, sample_dir, mod:model_textom, rec1d=False, sf_from_data=False, set_negative_zero=False ):
        """

        Parameters
        ----------
        dat : data class object or str
            if string it needs to be a 'fit'-pickle name
        mod : model class object
            only required if no pickle is loaded
        
        Attributes created
        ------------

        """
        print('Initializing Fit')
        datafile = 'data_1drec.h5' if rec1d else 'data_textom.h5'
        # import stuff from data and mod objects
        t0=time()
        with h5py.File( os.path.join(sample_dir,'analysis',datafile), 'r') as hf:
            self.data_unmasked = hf['data'][()].astype(data_type)
            self.scanmask = hf['scanmask'][()]
            self.gt_mask = hf['gt_mask'][()]
            if not rec1d:
                self.q_full = hf['q'][()]
                self.mask_detector = hf['mask_peaks'][()]
                
                # self.peak_reg = hf['peak_reg'][()]
                # self.hkl = [h.decode('utf-8') for h in hf['hkl']]
                # i'll take these 2 out of the txt file then it can be modified afterwards
                # and we load the peaks that are written inside
                path_peak_regions = os.path.join(sample_dir,'analysis','peak_regions.txt')
                self.peak_reg, self.hkl_full = load_peak_regions(path_peak_regions, exclude_empty=True)

            # check_version(hf['textom_version'][()])
        if set_negative_zero:
            self.data_unmasked[self.data_unmasked<0] = 0
        print('\tLoaded Data')

        self.mod = mod
        self.flag_use_peak = np.ones(len(self.hkl_full), bool)
        self.mask_data()
        self.list_peaks()
        self.sf_from_data = sf_from_data

        # self.mask_detector = np.genfromtxt(os.path.join(path,'analysis','fit_detmask.txt'),bool)
        
        # ######### do check if the test for changing detector angle direction
        # nD = self.data.shape[0]
        # dat_int = np.zeros( (nD,self.mask_detector.size), data_type )
        # # for k in range(self.data.shape[0]):
        # dat_int[:,self.mask_detector] = self.data
        # dat_int2 = dat_int.reshape((nD, *self.detShape))
        # self.mask_detector = np.flip(self.mask_detector)
        # dat_int3 = np.flip(dat_int2,axis=2).reshape(dat_int.shape)[:,self.mask_detector]
        # self.data = dat_int3        
        # #########

        self.kap_zero = np.where(mod.Kappa == 0)[0]
        ## initial guess for c0 from tomogram
        tomogram = mod.tomogram.flatten()/mod.tomogram.max() # normalize so c0 will be around 1
        tomogram[tomogram<0] = 0

        self.title, self.Beams, self.iBeams = mod.title, mod.Beams, mod.iBeams

        if not rec1d:
            ## renormalize the model
            g,t,_ = np.unravel_index(np.argmax(mod.Beams),mod.Beams.shape)
            iend = np.searchsorted( mod.iBeams[g, t, :],2**32-1 ) # for sparsearray
            T = sparsemult( 
                mod.Beams[g, t, :iend], 
                mod.iBeams[g, t, :iend], tomogram )
            # the simulated intensity will be multiplied by this to have C~O(1)
            gt_dat = np.argmin( np.linalg.norm(self.gt_mask - np.array([g,t]), axis=1) )
            if sf_from_data:
                self.sim_fac = 1 / T
            else:
                self.sim_fac = self.data[gt_dat].max()/mod.difflets[0].max() / T
            self.dat_scal = self.data[gt_dat].max()/T # and then do the rest in update_mod
            self.update_mod()

            # self.q = mod.Qq_det.reshape(mod.detShape)[:,0] # 1D q-vector
            # self.q = np.array([qr.mean() for qr in self.peak_reg])
            try:
                self.chi = mod.Chi_det.reshape(mod.detShape)[:,0]
            except:
                self.chi = mod.chi

        self.C = np.zeros((self.mod.x_p.shape[0], self.mod.difflets.shape[1]), data_type)
        self.choose_projections(info=False)

    def mask_data( self ):
        self.q = self.q_full[self.flag_use_peak]
        self.data = self.data_unmasked[:,:,self.flag_use_peak][:,self.mask_detector[:,self.flag_use_peak]]

    def toggle_peak_use( self, peak_idcs=[], flags=None ):
        if flags is not None:
            self.flag_use_peak = flags
        else:
            peak_idcs = np.atleast_1d(peak_idcs).astype(int)
            for p in peak_idcs:
                self.flag_use_peak[p] = not self.flag_use_peak[p]
        self.mask_data()
        self.update_mod()
        self.list_peaks()

    def list_peaks( self ):
        print('\tAvailable peaks:')
        print('\tIdx\tfit_flag\tq-range\t\thkl')
        for p in range(len(self.hkl_full)):
            print(f'\t{p}\t{self.flag_use_peak[p]}\t\t{self.peak_reg[p,0]:.2f}\t{self.peak_reg[p,1]:.2f}\t{self.hkl_full[p]}')

    def prepare_fit_c0( self, peak=None ):
        """ Calculates 0-dimensional data for fast estimation of the zero order 
        
        Parameters
        ----------
        peak : int or None
            index of the peak you want to use for 0-order fitting (should be as
            isotropic as possible), if None uses the whole dataset, default None
        """
        if peak:
            # not sure if this is still functional
            peak_start_idx = self.mask_detector[:peak*self.detShape[1]].sum()
            peak_end_idx = self.mask_detector[:(peak+1)*self.detShape[1]].sum()
            self.total_im = self.data[:,:,peak].sum(axis=1).sum(axis=1) # total peak intensity for each image
            self.data_0D = self.total_im / self.difflets_rot_masked[0,0,peak_start_idx:peak_end_idx].sum() 
        else:
            self.total_im = self.data.sum(axis=1) # total intensity for each image
            # self.total_proj = self.total_im.sum(axis=1) # total intensity for each projection
            self.data_0D = self.total_im / self.difflets_rot_masked[0,0].sum() 

    # def adjust_data_scaling( self ):
    #     """Reestimates the data based on the assumption that normalization constants
    #     contain noise. To be used after fitting the 0th order
    #     """
    #     print('\tAdjusting data scaling')
    #     self.prepare_fit_c0()
    #     # scal = self.insert_fov( self.projections_c0( self.C ) 
    #     #             ).sum(axis=1) / self.insert_fov( self.data_av ).sum(axis=1)
    #     # self.data_scal = np.tile(scal,(self.Beams.shape[1],1)).T[self.scanmask]
    #     # self.data = np.array( [s*d for s,d in zip(self.data_scal,self.data)])
    #     self.data_scal = self.projections_c0( self.C ) / self.data_0D
    #     self.data *= self.data_scal[:,np.newaxis]
    #     self.prepare_fit_c0()

    def projections_c0( self, C ):
        """ Wrapper for a numba-compiled function projecting the 0-order 
        coefficients and give a single value per pixel

        Parameters
        ----------
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index

        Return values
        ------------
        loss : float
        """
        proj_c0 = _projections_c0( C,
            self.gt_mask,
            self.Beams, self.iBeams, 
            )
        return proj_c0

    def set_orders( self, mod:model_textom, n_max, info=True, exclude_ghosts=True ):
        """ Changes the HSH order

        Calculates diffractlets and updates the fit object
        using functions in the model class

        Parameters
        ----------
        mod : model class object
        n_max : int
            highest HSH order for fitting
        """
        mod.set_orders( n_max, info=info, exclude_ghosts=exclude_ghosts )
        c_tmp = self.C
        self.update_mod_old(mod)
        if self.C.shape[1] >= c_tmp.shape[1]:
            self.C[:,:c_tmp.shape[1]] = c_tmp
        else:
            self.C = c_tmp[:,:self.C.shape[1]]

    def update_mod(self):
        """ Imports stuff from the model class

        Parameters
        ----------
        mod : model class object
        """
        # regroup diffractlets into peaks and mask them
        n_peaks = self.flag_use_peak.sum()
        self.hkl = np.array(self.hkl_full)[self.flag_use_peak]
        self.difflets_rot_masked = np.empty( (*self.mod.difflets.shape[:2], self.mask_detector[:,self.flag_use_peak].sum()), data_type )
        self.detShape = (self.mod.detShape[0], n_peaks) # effective detector shape after choosing peak regions
        self.odf_mode = self.mod.odf_mode
        self.odf_par = self.mod.odf_parameter
        if self.odf_mode == 'hsh':
            self.Sn = hsh.get_order_slices( self.odf_par, self.mod.symmetry )

        if self.sf_from_data:
            # take structure factor out of data:
            data_chimean = np.mean([masked_mean_axis0(d, self.mask_detector) for d in self.data_unmasked.reshape((self.data.shape[0],*self.detShape))], axis=0)

        for g, difflets_g in enumerate(self.mod.difflets):
            for l, dlet in enumerate(difflets_g):
                dlet_tmp = np.empty( self.detShape, data_type)
                for k in range(n_peaks):
                    # if hex_notation and self.mod.symmetry[0] == '6':
                    #     hkls = np.array([miller_to_hex(hkl) for hkl in self.mod.hkl])
                    # else:
                    #     hkls = self.mod.hkl
                    hkls = np.column_stack([ # this should be safe for miller and hex notation
                        self.mod.hkl[:,0],self.mod.hkl[:,1],self.mod.hkl[:,-1]
                        ])

                    # check which peaks are specified in the file for this data bin
                    coords = re.findall(r'\[([-?\d, ]+)\]', self.hkl[k])
                    matches = []
                    for c in coords:
                        vals = np.fromstring(c, sep=',', dtype=int)
                        vals = np.array([vals[0], vals[1], vals[-1]])
                        # find where in myarray this triplet appears
                        idx = np.where((hkls == vals).all(axis=1))[0]
                        # matches.extend(idx.tolist())
                        matches.append(idx)
                    mask_peaks = np.zeros(self.mod.hkl.shape[0], bool)
                    mask_peaks[matches] = True

                    dlet_tmp[:,k] = dlet[:,mask_peaks].sum(axis=1) # sum over all peaks that are in the bin

                    if self.sf_from_data:
                        dlet_tmp[:,k] *= data_chimean[k] / self.mod.structure_factors[mask_peaks].sum()
                # try:
                self.difflets_rot_masked[g,l] = dlet_tmp[self.mask_detector[:,self.flag_use_peak]]
                    # self.difflets_rot_masked[g,l] = dlet_tmp[self.mask_detector]
                # except:
                #     raise dimension_error('Check if No of peak regions x No of azimuthal bins agree with the detector mask size')
        self.difflets_rot_masked *= self.dat_scal / self.difflets_rot_masked[0].max()

    def choose_projections(self, mode='full', info=True):
        """ Decides which sample rotations to use for fitting

        Parameters
        ----------
        mode : str
        """
        # Choose projections:
        Ng0 = self.Beams.shape[0] # total number of rotations
        if mode=='notilt': # only kappa = 0
            self.chosen_proj = self.kap_zero
        elif mode=='half': # every second
            self.chosen_proj = range(0,Ng0,2)
        elif mode=='third': # every third
            self.chosen_proj = range(0,Ng0,3)
        elif mode=='asnotilt': # equidistant, as many as at kappa = 0
            n = self.kap_zero.size
            self.chosen_proj = np.arange(0,Ng0,int(Ng0/n))[:n]
        else: # all of them
            self.chosen_proj = range(Ng0)
        tmp_scanmask = self.scanmask.copy()
        mask = np.full_like(tmp_scanmask, False)
        mask[self.chosen_proj,:]=True
        tmp_scanmask[~mask] = False
        self.gt_mask = np.transpose(np.where(tmp_scanmask))
        if info:
            print('Choose projections: %s' % mode)

    def loss( self, C ):
        """ Wrapper for a numba-compiled function calculating the loss function

        Parameters
        ----------
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index

        Return values
        ------------
        loss : float
        """
        loss = _loss( C,
            self.data, self.gt_mask,
            self.Beams, self.iBeams, 
            self.difflets_rot_masked,
            )
        return loss

    def loss_c0( self, C ):
        """ Wrapper for a numba-compiled function calculating the loss function
        using only the zero-order coefficients compared to the averaged images

        Parameters
        ----------
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index

        Return values
        ------------
        loss : float
        """
        loss = _loss_c0( C,
            self.data_0D, self.gt_mask,
            self.Beams, self.iBeams, 
            )
        return loss
    
    def grad_full( self, C ):
        """ Wrapper for a numba-compiled function calculating the gradient
        of the loss function for all sHSH coefficients

        Parameters
        ----------
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index

        Return values
        ------------
        grad : 2D ndarray, float
            gradient for all the sHSH coefficients
            dim: 0: voxel index, 1: sHSH index
        """
        grad = _grad_full( C,
            self.data, self.gt_mask,
            self.Beams, self.iBeams, 
            self.difflets_rot_masked,
            )
        return grad

    def grad_highest( self, C ):
        """ Wrapper for a numba-compiled function calculating the gradient
        of the loss function for the highest order sHSH coefficients 

        Parameters
        ----------
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index

        Return values
        ------------
        grad : 2D ndarray, float
            gradient for all the sHSH coefficients
            dim: 0: voxel index, 1: sHSH index
        """            
        grad = _grad_highest( C,self.Sn[-1],   
            self.data, self.gt_mask,
            self.Beams, self.iBeams, 
            self.difflets_rot_masked,
            )
        return grad
    
    def grad_allbut0( self, C ):
        """ Wrapper for a numba-compiled function calculating the gradient
        of the loss function for all but 0-order sHSH coefficients 

        Parameters
        ----------
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index

        Return values
        ------------
        grad : 2D ndarray, float
            gradient for all the sHSH coefficients
            dim: 0: voxel index, 1: sHSH index
        """            
        grad = _grad_allbutC0( C,self.Sn[1:],   
            self.data, self.gt_mask,
            self.Beams, self.iBeams, 
            self.difflets_rot_masked,
            )
        return grad
    
    def grad_c0( self, C ):
        """ Wrapper for a numba-compiled function calculating the gradient
        of the loss function for the zero order sHSH coefficients from 
        the average image value

        Parameters
        ----------
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index

        Return values
        ------------
        grad : 2D ndarray, float
            gradient for all the sHSH coefficients
            dim: 0: voxel index, 1: sHSH index
        """
        grad = _grad_c0( C,        
            self.data_0D, self.gt_mask,
            self.Beams, self.iBeams, 
            )
        return grad

    def lipschitz( self ):
        """ Wrapper for a numba-compiled function calculating the Lipschitz 
        constant for the zero order sHSH coefficients

        Return values
        ------------
        Lipschitz : 1D ndarray, float
        """
        Lipschitz = _lipschitz_c0( self.C,
            self.gt_mask,
            self.Beams, self.iBeams,
            )
        return Lipschitz   

    def projection( self, g, C, info=True ):
        """ Wrapper for a numba-compiled function calculating a series
        of diffraction pattern corresponding to one sample orientation

        Parameters
        ----------
        g : int
            rotation index
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index

        Return values
        ------------
        images : 2D ndarray, float
            dim 0: translation, 1: detector points
        """
        if info:
            print('\tCalculating projection %d' % g )
        images = _projection(
            C, g, self.scanmask,
            self.Beams, self.iBeams, 
            self.difflets_rot_masked,
        )
        return images

    def image( self, C, g, t ):
        """Simulates a single diffraction image for rotation g and 
        translation t

        Parameters
        ----------
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index
        g : int
            rotation index
        t : int
            translation index
        Return values
        ------------
        image : 1D ndarray, float
        """
        iend = np.searchsorted(self.iBeams[g,t,:],2**32-1) # for sparsearray
        # get projected coefficients for the corresponding beam rotation/translation
        c_proj = integrate_c( self.Beams[g,t,:iend], self.iBeams[g,t,:iend], C )
        # calculate the image
        image = c_proj @ self.difflets_rot_masked[g] 
        # image = c_proj @ self.difflets_rot[g] 
        return image

    def MDL( self, loss ):
        """ Calculates the Medium description length for order selection
        
        M. Hansen and B. Yu, ?Model selection and the principle of minimum 
        description length, Journal of the American Statistical Association,
        vol. 96, no. 454, pp. 746-774, 2001.

        Parameters
        ----------
        loss : float

        Return values
        ------------
        MDL : float
        """
        # Calculate the number of datapoints
        # lowmask = np.array(
        #     [beam[0]<2**23-1 for beam_g in self.iBeams for beam in beam_g])
        # im_mask = np.logical_and( lowmask, self.scanmask.flatten() )            
        #     [beam[0]<2**23-1 for beam_g in self.iBeams[self.chosen_proj] for beam in beam_g])
        # im_mask = np.logical_and( lowmask, self.scanmask[self.chosen_proj].flatten() )
        N_images = np.sum( self.scanmask )
        N_data = N_images * self.mask_detector[:,self.flag_use_peak].sum()
        # the number of parameters
        n_par = self.C[self.C[:,0]>0].size
        # calculate the medium description length
        MDL = N_data/2 * np.log(loss) + n_par/2 * np.log(N_data)
        return MDL

    def insert_fov( self, data ):
        d_out = np.zeros_like( self.scanmask, data_type )
        d_out[self.scanmask] = data
        return d_out
    
    def insert_fov_1d( self, data ):
        d_out = np.zeros( (*self.scanmask.shape, data.shape[1]), data_type )
        d_out[self.scanmask] = data
        return d_out
    
    def insert_images( self, images_masked ):
        images = np.zeros((images_masked.shape[0],*self.detShape), images_masked.dtype)
        images[:,self.mask_detector[:,self.flag_use_peak]] = images_masked
        return images
    
    def check_orientations( self, g ):

        proj = np.zeros( (self.Beams.shape[1], *self.mask_detector[:,self.flag_use_peak].shape), data_type )
        proj[:,self.mask_detector[:,self.flag_use_peak]] = self.projection( g, self.C, info=False )
        dat = np.zeros_like(proj)
        d_mskd = self.data[ np.where(self.gt_mask[:,0]==g)[0] ]
        for nt, t in enumerate(self.gt_mask[self.gt_mask[:,0]==g][:,1]):
            dat[t,self.mask_detector[:,self.flag_use_peak]] = d_mskd[nt]

        ori_dat = np.empty( (self.q.size, self.Beams.shape[1]), data_type )
        ori_fit = ori_dat.copy()
        for i_peak, q in enumerate(self.q):
            chi_hlf = self.chi[:int(self.chi.size/2)]
            ori_dat[i_peak] = _get_mean_ori( dat, self.scanmask[g], i_peak, chi_hlf )
            ori_fit[i_peak] = _get_mean_ori( proj, self.scanmask[g], i_peak, chi_hlf )

        ori_dat[ori_dat==0]=np.nan
        ori_fit[ori_fit==0]=np.nan
        return ori_dat, ori_fit

    def projection_residuals( self, g ):
        proj = self.projection( g, self.C, info=False )
        dat = self.data[ np.where(self.gt_mask[:,0]==g)[0] ]
        res = np.zeros( proj.shape[0], data_type )
        res[self.scanmask[g]] = (( proj[self.scanmask[g]] - dat )**2 ).sum(axis=1)
        return res

@njit(parallel=True)
def _loss( C, data, gt_mask, Beams, iBeams, difflets_rot ): 
    """ Calculates the loss function as the sum of squared residuals

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    data : 3D ndarray, float
        dim: 0: rotation, 1: translation, 2: detector points
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index
    difflets_rot : 3D ndarray, float
        diffractlets for all sample rotations
        dim: 0: sample rotation, 1: sHSH indices, 2: detector points

    returns: 
    ------------
    loss : float
    """
    loss = 0.0
    for k in prange(gt_mask.shape[0]):
        g,t = gt_mask[k,0], gt_mask[k,1]
        iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
        # get projected coefficients for the corresponding beam rotation/translation
        c_proj = integrate_c( Beams[g,t,:iend], iBeams[g,t,:iend], C )
        # calculate the image
        I_sim = c_proj @ difflets_rot[g] 
        # calculate the squared sum of residuals
        residuals = data[k] - I_sim 
        loss += (residuals**2).sum()# / (data[k]**2).sum()
    return loss

@njit(parallel=True)
def _loss_c0( C, data_av, gt_mask, Beams, iBeams ): 
    """ Calculates the loss function as the sum of squared residuals
    using only the zero-order coefficients compared to the averaged images

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    data_av : 2D ndarray, float
        dim: 0: rotation, 1: translation
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index

    returns: 
    ------------
    loss : float
    """
    loss = 0.0
    for k in prange(gt_mask.shape[0]):
        g,t = gt_mask[k,0], gt_mask[k,1]
        iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
        # get projected coefficients for the corresponding beam rotation/translation
        c0_proj = sparsemult( Beams[g,t,:iend], iBeams[g,t,:iend], C[:,0] )
        loss += (data_av[k] - c0_proj)**2# / (data_av[k])**2 
    return loss

@njit(parallel=True)
def _projections_c0( C, gt_mask, Beams, iBeams ): 
    """ Calculates the loss function as the sum of squared residuals
    using only the zero-order coefficients compared to the averaged images

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index

    returns: 
    ------------
    loss : float
    """
    c0_proj = np.empty(gt_mask.shape[0], data_type)
    for k in prange(gt_mask.shape[0]):
        g,t = gt_mask[k,0], gt_mask[k,1]
        iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
        # get projected coefficients for the corresponding beam rotation/translation
        c0_proj[k] = sparsemult( Beams[g,t,:iend], iBeams[g,t,:iend], C[:,0] )
    return c0_proj

@njit(parallel=True)
def _grad_full( C, data, gt_mask, Beams, iBeams, difflets_rot ): 
    """ Calculates the gradient for all coefficients

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    data : 3D ndarray, float
        dim: 0: rotation, 1: translation, 2: detector points
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index
    difflets_rot : 3D ndarray, float
        diffractlets for all sample rotations
        dim: 0: sample rotation, 1: sHSH indices, 2: detector points

    Return values
    ------------
    grad : 2D ndarray, float
        gradient for all the sHSH coefficients
        dim: 0: voxel index, 1: sHSH index
    """
    grad = np.zeros_like( C )
    for k in prange(gt_mask.shape[0]):
        g,t = gt_mask[k,0], gt_mask[k,1]
        iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
        # get projected coefficients for the corresponding beam rotation/translation
        c_proj = integrate_c( Beams[g,t,:iend], iBeams[g,t,:iend], C )
        # calculate the image
        I_sim = c_proj @ difflets_rot[g] 
        # get the gradient on the coefficients from this projection
        residuals = data[k] - I_sim 
        grad_proj =  - 2 * difflets_rot[g] @ residuals
        # put the beam-intensity weighted value into each voxel
        for isp in range(iend): # sparse index
            iv = iBeams[g,t,isp] # voxel index
            grad[iv] += Beams[g,t,isp] * grad_proj
    return grad

@njit(parallel=True)
def _grad_highest(C, order_slice, data, gt_mask, Beams, iBeams, difflets_rot ): 
    """ Calculates the gradient for the highest order coefficients

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    data : 3D ndarray, float
        dim: 0: rotation, 1: translation, 2: detector points
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index
    difflets_rot : 3D ndarray, float
        diffractlets for all sample rotations
        dim: 0: sample rotation, 1: sHSH indices, 2: detector points

    Return values
    ------------
    grad : 2D ndarray, float
        gradient for all the sHSH coefficients
        dim: 0: voxel index, 1: sHSH index
    """
    grad = np.zeros_like( C )
    for k in prange(gt_mask.shape[0]):
        g,t = gt_mask[k,0], gt_mask[k,1]
        iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
        # get projected coefficients for the corresponding beam rotation/translation
        c_proj = integrate_c( Beams[g,t,:iend], iBeams[g,t,:iend], C )
        # calculate the image
        I_sim = c_proj @ difflets_rot[g] 
        # get the gradient on the coefficients from this projection
        residuals = data[k] - I_sim 
        grad_proj =  - 2 * difflets_rot[g, order_slice[0]:order_slice[1]] @ residuals
        # put the beam-intensity weighted value into each voxel
        for isp in range(iend): # sparse index
            iv = iBeams[g,t,isp] # voxel index
            grad[iv, order_slice[0]:order_slice[1]] += Beams[g,t,isp] * grad_proj
    return grad

@njit(parallel=True)
def _grad_allbutC0(C, order_slice, data, gt_mask, Beams, iBeams, difflets_rot ): 
    """ Calculates the gradient for the highest order coefficients

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    data : 3D ndarray, float
        dim: 0: rotation, 1: translation, 2: detector points
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index
    difflets_rot : 3D ndarray, float
        diffractlets for all sample rotations
        dim: 0: sample rotation, 1: sHSH indices, 2: detector points

    Return values
    ------------
    grad : 2D ndarray, float
        gradient for all the sHSH coefficients
        dim: 0: voxel index, 1: sHSH index
    """
    grad = np.zeros_like( C )
    for k in prange(gt_mask.shape[0]):
        g,t = gt_mask[k,0], gt_mask[k,1]
        iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
        # get projected coefficients for the corresponding beam rotation/translation
        c_proj = integrate_c( Beams[g,t,:iend], iBeams[g,t,:iend], C )
        # calculate the image
        I_sim = c_proj @ difflets_rot[g] 
        # get the gradient on the coefficients from this projection
        residuals = data[k] - I_sim 
        grad_proj =  - 2 * difflets_rot[g, 1:] @ residuals
        # put the beam-intensity weighted value into each voxel
        for isp in range(iend): # sparse index
            iv = iBeams[g,t,isp] # voxel index
            grad[iv, 1:] += Beams[g,t,isp] * grad_proj
    return grad

@njit(parallel=True)
def _grad_c0( C, data_av, gt_mask, Beams, iBeams ): 
    """ Calculates the gradient for the zero order coefficients
    from the averaged images

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    data_av : 2D ndarray, float
        dim: 0: rotation, 1: translation
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index

    returns: 
    ------------
    grad : 2D ndarray, float
        gradient for all the sHSH coefficients
        dim: 0: voxel index, 1: sHSH index
    """
    grad = np.zeros_like( C )
    for k in prange(gt_mask.shape[0]):
        g,t = gt_mask[k,0], gt_mask[k,1]
        iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
        # get projected coefficients for the corresponding beam rotation/translation
        c0_proj = sparsemult( Beams[g,t,:iend], iBeams[g,t,:iend], C[:,0] )
        residual = data_av[k] - c0_proj
        # assign gradient to the affected voxels
        grad[ iBeams[g,t,:iend], 0 ] += - 2 * residual * Beams[g,t,:iend]
    return grad

@njit(parallel=True)
def _lipschitz_c0( C, gt_mask, Beams, iBeams ): 
    """ Calculates the Lipschitz constant for the zero order coefficients

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index

    returns: 
    ------------
    L : 1D ndarray, float
    """
    L = np.zeros_like( C[:,0], data_type)
    for k in prange(gt_mask.shape[0]):
        g,t = gt_mask[k,0], gt_mask[k,1]
        iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
        # assign contribution to the affected voxels
        L[ iBeams[g,t,:iend] ] += Beams[g,t,:iend]#**2
    return L

@njit(parallel=True)
def _projection( C, g, scanmask, Beams, iBeams, difflets_rot ):
    """Simulates diffraction images from a 2D scan over the sample

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxelN
        dim: 0: voxel index, 1: sHSH index
    g : int
        rotation index
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index
    difflets_rot : 3D ndarray, float
        diffractlets for all sample rotations
        dim: 0: sample rotation, 1: sHSH indices, 2: detector points
        
    Return values
    ------------
    images : 3D ndarray, float
        array of resulting scattering intensity for each point on the 
        detector, for each rotation and translation
        dimensions: 0: rotation, 1: translation, 2: detector points
    """
    images = np.zeros( (Beams.shape[1], difflets_rot.shape[2]), data_type )
    for t in prange(Beams.shape[1]): # translations
        if scanmask[g,t]:
            iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
            # get projected coefficients for the corresponding beam rotation/translation
            c_proj = integrate_c( Beams[g,t,:iend], iBeams[g,t,:iend], C ).astype(data_type)
            # calculate the image
            images[t] = c_proj @ difflets_rot[g] 
    return images

@njit()
def _rotate_difflets_stack( Rs_stack, difflets ):
    ''' calculates the symmetrized HSH rotation matrices and the corresponding
    diffraction images for all sHSHs of order n and all rotations Gs
    '''
    difflets_stack = np.zeros( (Rs_stack.shape[0], difflets.shape[0], difflets.shape[1]), data_type )
    for g, Rs_g in enumerate(Rs_stack):
        difflets_stack[g] = Rs_g @ difflets
    
    return difflets_stack

@njit(parallel=True)
def _get_mean_ori( im, scanmask, i_peak, chi ):
    ori = np.zeros( im.shape[0], data_type )
    for t in prange(im.shape[0]):
        if scanmask[t]:
            im2D = im[t]#.reshape( detShape )
            # get azimuthal data for one peak
            I_azi = im2D[:,i_peak]
            nphi_2 = int( I_azi.size/2 ) 
            # divide data to from 0 to pi and pi to 2pi
            I_azi1, I_azi2 = I_azi[nphi_2:], I_azi[:nphi_2]
            # remove mask/detector gaps
            I_azi1[I_azi1==0] = I_azi2[I_azi1==0]
            I_azi2[I_azi2==0] = I_azi1[I_azi2==0]
            # average both parts
            I_azi_folded = (I_azi1 + I_azi2)/2
            # to calculate the mean
            x_mean = ( I_azi_folded * np.cos( 2*chi ) ).sum() / I_azi_folded.sum()
            y_mean = ( I_azi_folded * np.sin( 2*chi ) ).sum() / I_azi_folded.sum()
            chi_mean = np.arctan2( y_mean, x_mean ) / 2 + np.pi/2

            ori[t] = chi_mean
    return ori

# not used
# @njit(parallel=True)
# def _get_powder_1D( data, mask_detector, detShape, iBeams, scanmask ):
#     data_1D = np.zeros((detShape[0]),data_type)
#     for g in prange(iBeams.shape[0]): # rotations
#         for t in range(iBeams.shape[1]): # translations
#             if ( scanmask[g,t] and (iBeams[g,t,0] < 2**32-1) ):
#                 data_1D += _masked_to_1D(
#                     data[g,t,mask_detector], mask_detector, detShape)
#     return data_1D

# @njit
# def _masked_to_1D( A, mask_detector, detShape ):
#     A_full = np.zeros( detShape, data_type ).flatten()
#     A_full[mask_detector] = A
#     A_1D = np.array([dq[dq>0].mean() for dq in A_full.reshape(detShape)])
#     return A_1D

@njit(parallel=True)
def beamnorm(gt_mask, Beams ): 
    Nb = np.empty( gt_mask.shape[0] )
    for k in prange(gt_mask.shape[0]):
        g,t = gt_mask[k,0], gt_mask[k,1]
        Nb[k] = Beams[g,t].sum()
    return Nb


import sys
from time import time
from matplotlib.colors import to_rgb
from matplotlib.lines import Line2D
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button
from matplotlib.animation import FuncAnimation
import numpy as np

from orix import plot
# from orix.quaternion import Orientation, Misorientation, Rotation
from orix.vector import Vector3d, AxAngle
from orix.quaternion import Orientation, OrientationRegion, Rotation, Misorientation
import orix.quaternion.symmetry as osym #import C1, Oh, D2
from orix.crystal_map import Phase
from orix.vector import Miller, Vector3d

# from .model import model
from ..model import symmetries as sym
from ..model import rotation as rot
from ..misc import crop_to_non_nan_region, smallest_subvolume
from ...config import data_type

############################################################################################+
    
# def plot_odf( orientations, odf, symmetry, num_samples=1000, title='' ):

#     # draw samples out of the odf
#     cdf = np.cumsum( odf/odf.sum() )

#     # Draw random samples from a uniform distribution
#     random_numbers = np.random.rand(int(num_samples))

#     # Find the corresponding indices in the CDF
#     indices = np.searchsorted(cdf, random_numbers)
#     indices[indices==odf.size] = 0

#     # make them quaternions
#     quat = rot.QfromOTP( orientations[indices] )
#     # sample_r = Rotation([quat])
#     # sample = Orientation(sample_r)
#     # sample.symmetry = getattr( osym, sym.get_SFnotation( symmetry ) )

#     # # plots a 3D odf in the fundamental zone
#     # sample.scatter()
#     plot_points_in_fz( quat, symmetry )
#     plt.title(title)
#     plt.show()

def plot_odf( unit_quaternions, odf, symmetry, num_samples=1000, title='' ):
    if odf.sum() == 0:
        print('ODF normalization equals 0, provide other coefficents')
    else:
        # draw samples out of the odf
        cdf = np.cumsum( odf/odf.sum() )

        # Draw random samples from a uniform distribution
        random_numbers = np.random.rand(int(num_samples))

        # Find the corresponding indices in the CDF
        indices = np.searchsorted(cdf, random_numbers)
        indices[indices==odf.size] = 0

        # # plots a 3D odf in the fundamental zone
        plot_points_in_fz( unit_quaternions[indices], symmetry, title=title )

def plot_points_in_fz( unit_quaterions, symmetry, title='' ):
    # make them quaternions
    # quat = rot.QfromOTP( orientations )
    sample_r = Rotation([unit_quaterions])
    sample = Orientation(sample_r)
    sample.symmetry = getattr( osym, sym.get_SFnotation( symmetry ) )

    # plots a 3D odf in the fundamental zone
    sample.scatter()

    plt.title(title)

############################################################################################+
# from orix.quaternion import Orientation, OrientationRegion, Rotation, Misorientation
# import orix.quaternion.symmetry as osym #import C1, Oh, D2
# # plot a single piont in a unit cell with vector pointing to it
# axan = np.array([[np.pi/2,np.pi/4,np.pi/4]])
# # f,ax = plt.subplots(4)
# symmetries = ['2','222','6','432']#
# k=1
# for symmetry in symmetries:
#     quat = rot.QfromOTP(axan)
#     axan_vec = quat[0,1:]/np.linalg.norm(quat[0,1:]) * axan[0,0]
#     oror = Orientation(Rotation(quat))
#     oror.symmetry = getattr( osym, sym.get_SFnotation(symmetry))
#     # f.add_axes([0.1, 0.1, 0.8, 0.8])
#     oror.scatter()#figure=f,position=(4,1,k))
#     k+=1
#     # plt.quiver(0,0,0,*axan_vec,color='r')
############################################################################################+

def inverse_pf_map( g, ppgrp ):
    """"""
    orix_pg = getattr( osym, sym.get_SFnotation( ppgrp ) )
    ori = make_ori(g,orix_pg)

    ckey = plot.IPFColorKeyTSL(orix_pg)

    directions = [(1, 0, 0), (0, 1, 0), (0, 0, 1)]
    titles = ["X", "Y", "Z"]

    fig, axes = plt.subplots( 2,2, figsize=(15, 10))
    for i, ax in enumerate(axes.flatten()[:3]):
        ckey.direction = Vector3d(directions[i])
        # Invert because orix assumes lab2crystal when coloring orientations
        ax.imshow(ckey.orientation2color(~ori).transpose(1,0,2))
        ax.set_title(f"IPF-{titles[i]}")
        ax.axis("off")

    # Add color key
    ax_ipfkey = plt.subplot( 224,
        # [0.932, 0.37, 0.1, 0.1],  # (Left, bottom, width, height)
        projection="ipf",
        symmetry=ori.symmetry.laue,
    )
    ax_ipfkey.plot_ipf_color_key()
    ax_ipfkey.set_title("")
    # axes[1,1].set_projection = "ipf"
    # axes[1,1].symmetry=ori.symmetry.laue
    # axes[1,1].plot_ipf_color_key()
    # axes[1,1].set_title("")
    fig.subplots_adjust(wspace=0.01)

def inverse_pf_stack( g, symmetry, mask_voxels ):
    """Creates an inverse pole figure map of every slice of a 

    Parameters
    ----------
    g : ndarray
        orientations in OTP notation
    ppgrp : str
        proper point group

    Returns
    -------
    list
        for x/y/z: 3D inv polefigures
    """
    orix_pg = getattr( osym, sym.get_SFnotation( symmetry ) )
    ori = make_ori( g.reshape((np.prod(g.shape[:-1]),g.shape[-1])), orix_pg )

    ckey = plot.IPFColorKeyTSL(orix_pg)

    directions = [(1, 0, 0), (0, 1, 0), (0, 0, 1)]
    
    col_stack = []
    for k in range(len(directions)):
        ckey.direction = Vector3d(directions[k])
        colors = np.zeros((np.prod(g.shape[:-1]),3), data_type)
        # Invert because orix assumes lab2crystal when coloring orientations
        colors[mask_voxels] = ckey.orientation2color(~ori[mask_voxels])
        col_stack.append(colors.reshape((*g.shape[:-1],g.shape[-1])))

    return np.array(col_stack)#.transpose((1,2,3,0,4))#col_stack

def plot_ipf_movie( data, ppgrp, sliceplane='z', save_vid_path=False, fps=5, show=True ):

    if sliceplane == 'x':
        xl,yl = 'y','z' # labels for plot
    elif sliceplane == 'y':
        data=data.transpose(0,2,1,3,4)
        xl,yl = 'x','z' # labels for plot
    elif sliceplane == 'z':
        data=data.transpose(0,3,1,2,4)
        xl,yl = 'x','y' # labels for plot
    else:
        print('\tsliceplane not recognized')
        return
    
    data = np.array([smallest_subvolume(d)[0] for d in data])
    num_images = data.shape[1]
    # Initial dataset and image index
    current_image_index = 0

    fig, axes = plt.subplots( 2,2, figsize=(15, 10))
    fig.delaxes(axes[1,1])
    titles = ["X", "Y", "Z"]
    imgs_display = []
    for i, ax in enumerate(axes.flatten()[:3]):
        # Invert because orix assumes lab2crystal when coloring orientations
        imgs_display.append(ax.matshow(data[i,data.shape[1]//2].transpose(1,0,2)))
        ax.set_title(f"IPF-{titles[i]}")
        ax.axis("off")
    orix_pg = getattr( osym, sym.get_SFnotation( ppgrp ) )
    # Add color key
    ax_ipfkey = plt.subplot( 224,
        # [0.932, 0.37, 0.1, 0.1],  # (Left, bottom, width, height)
        projection="ipf",
        symmetry=orix_pg.laue,
    )
    ax_ipfkey.plot_ipf_color_key()
    ax_ipfkey.set_title("")
    # Create a slider for switching between images within a dataset
    ax_image_slider = plt.axes([0.25, 0.03, 0.65, 0.03], facecolor='lightgoldenrodyellow')
    image_slider = Slider(ax_image_slider, sliceplane, 0, num_images - 1, valinit=0, valstep=1)
    def update(val):
        current_image_index = int(image_slider.val)  # Get the image index from the slider
        current_image = data[:,current_image_index]
        for i, ax in enumerate(axes.flatten()[:3]):
            imgs_display[i].set_data( current_image[i].transpose(1,0,2) )  # Update the image data
        fig.canvas.draw_idle()  # Redraw the figure

    # Attach the update functions to the sliders
    image_slider.on_changed(update)

    # Add a button for playing the video
    ax_button = plt.axes([0.1, 0.03, 0.1, 0.04])
    button = Button(ax_button, 'Play')

    playing = [False]  # Use a mutable object to toggle playing state

    def toggle_play(event):
        playing[0] = not playing[0]  # Toggle state
        if playing[0]:
            button.label.set_text("Pause")
            ani.event_source.start()
        else:
            button.label.set_text("Play")
            ani.event_source.stop()

    button.on_clicked(toggle_play)

    # Automatic playback with animation
    def update_frame(i):
        if playing[0]:
            next_frame = (int(image_slider.val) + 1) % num_images
            image_slider.set_val(next_frame)

    ani = FuncAnimation(fig, update_frame, frames=num_images, interval=100)  # Interval in milliseconds

    if save_vid_path:
        toggle_play(None)
        # Save the animation as an MP4 file
        ani.save(save_vid_path, writer='pillow', fps=fps)
        toggle_play(None)

    if show:
        plt.show()

def plot_pole_figure( G_odf, odf, symmetry, hkl=(1,0,0), 
                     hemisphere = 'upper',
                     mode='image', title='',
                     num_samples=1e4, alpha=0.05 ):

    # draw samples out of it
    cdf = np.cumsum( odf/odf.sum() )

    # Draw random samples from a uniform distribution
    random_numbers = np.random.rand(int(num_samples))

    # Find the corresponding indices in the CDF
    indices = np.searchsorted(cdf, random_numbers)-1

    # make them quaternions
    quat = rot.QfromOTP( G_odf[indices] )
    sample_r = Rotation([quat])
    sample = Orientation(sample_r)
    sample.symmetry = getattr( osym, sym.get_SFnotation( symmetry ) )

    g = Miller(hkl=hkl, phase=Phase(point_group=sample.symmetry))
    g = g.symmetrise(unique=True)
    poles = sample.inv().outer(g, lazy=True, progressbar=True, chunk_size=2000)

    plt.rcParams.update(
        {
            "figure.figsize": (6, 5),
            "lines.markersize": 2,
            "font.size": 15,
            "axes.grid": False,
        }
    )
    w, h = plt.rcParams["figure.figsize"]
    if mode=='scatter':
        poles.scatter(
            hemisphere=hemisphere,
            alpha=alpha,
            figure_kwargs={"figsize": (2 * h, h)},
            axes_labels=["X", "Y"],
        )
    else:
        poles.pole_density_function(
            hemisphere=hemisphere, log=False, figure_kwargs={"figsize": (2 * h, h)}
        )
    plt.suptitle( title )

def make_ori( g, sym ):
    # print('\tSetting up orix orientations')
    g_flat = g.reshape((np.prod(g.shape[:-1]),g.shape[-1])) # flatten orientations
    quat = rot.QfromOTP( g_flat ) # make quaternions
    ori_rot = Rotation([quat]) # convert to orix rotations
    ori = Orientation(ori_rot) # convert to orix orientations
    ori = ori.reshape((g.shape[:-1])) # bring back to original shape
    ori.symmetry = sym # apply symmetry
    ori = ori.map_into_symmetry_reduced_zone() # check if this is necessary
    return ori

def test_mori(g, ppgrp):
    orix_pg = getattr( osym, sym.get_SFnotation( ppgrp ) )
    ori = make_ori(g, orix_pg).flatten()
    g_fl = g.reshape((np.prod(g.shape[:-1]),g.shape[-1]))
    msk = ~np.isnan(g_fl[:,0])

    ori = ori[msk]
    g_fl = g_fl[msk]

    t0=time()
    mori_all = Misorientation(~ori[:-1] * ori[1:], symmetry=(orix_pg,orix_pg))
    mori_all.map_into_symmetry_reduced_zone()
    dis_mori = mori_all.angle
    print(time()-t0)
    # t0=time() # this takes much longer
    # for k in range(ori.size-1):
    #     mori_single = Misorientation(~ori[k] * ori[k+1], symmetry=(orix_pg,orix_pg))
    #     mori_single.map_into_symmetry_reduced_zone()
    #     mori_single.angle
    # print(time()-t0)
    t0=time()
    gen = sym.generators(ppgrp)
    dis = rot.ang_distance(rot.QfromOTP(g_fl[:-1]),rot.QfromOTP(g_fl[1:]),gen)
    dis_mori_rot = rot.misorientation(rot.QfromOTP(g_fl[:-1]),rot.QfromOTP(g_fl[1:]),gen)
    print(time()-t0)



import numpy as np
## Define diffraction-related parameters:
# x-ray energy in keV
E_keV = 15.2
# q range for fitting: (lower,upper) boundary in nm^-1
q_range = (10,35)
# path to crystal cif file
cifPath = 'analysis/BaCO3.cif'
# parameters for diffractlet calculation
cutoff_structure_factor=1e-4
max_hkl=6

odf_mode = 'hsh' # 'grid' # 
hsh_max_order = 12 # ignored if odf_mode is grid
grid_resolution = 15 # degree # ignored if odf_mode is hsh

# options
use_structure_factors_from_data = False
use_upgraded_point_group = True